#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : test_learning_rates.py                                            #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Sunday, May 17th 2020, 2:27:44 am                           #
# Last Modified : Sunday, May 17th 2020, 2:34:01 am                           #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Tests learning rate classes."""
#%%
import math
import os
from pathlib import Path
import sys
testdir = str(Path(__file__).parents[2])
testdatadir = os.path.join(testdir, 'test_data')
sys.path.append(testdatadir)

import pandas as pd
import pytest
from pytest import mark
import numpy as np

from mlstudio.supervised.observers.learning_rate import StepDecay, TimeDecay
from mlstudio.supervised.observers.learning_rate import SqrtTimeDecay
from mlstudio.supervised.observers.learning_rate import ExponentialDecay, PolynomialDecay
from mlstudio.supervised.observers.learning_rate import PolynomialStepDecay
from mlstudio.supervised.observers.learning_rate import ExponentialStepDecay, PowerSchedule
from mlstudio.supervised.observers.learning_rate import BottouSchedule, Improvement

@mark.observer
@mark.lrs
@mark.step_decay
class StepDecayTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_step_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = StepDecay(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = StepDecay(eta0=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = StepDecay(eta0=1)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = StepDecay(eta_min=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = StepDecay(eta_min=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = StepDecay(eta_min=1)
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = StepDecay(decay_factor=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = StepDecay(decay_factor=-1)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = StepDecay(decay_factor=2)
            lrs.on_train_begin()       
        # Validate decay_steps
        with pytest.raises(TypeError):
            lrs = StepDecay(decay_steps=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = StepDecay(decay_steps=-1)
            lrs.on_train_begin()
        with pytest.raises(TypeError):
            lrs = StepDecay(decay_steps=np.inf)
            lrs.on_train_begin()      

    def test_step_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_step_decay.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = StepDecay(eta0=0.1, eta_min=0.01,
                        decay_factor=0.5, decay_steps=5)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg

@mark.observer
@mark.lrs
@mark.time_decay
class TimeDecayTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_time_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = TimeDecay(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = TimeDecay(eta0=0, decay_factor=0.5)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = TimeDecay(eta0=1, decay_factor=0.5)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = TimeDecay(eta_min=None, decay_factor=0.5)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = TimeDecay(eta_min=0, decay_factor=0.5)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = TimeDecay(eta_min=1, decay_factor=0.5)
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = TimeDecay(decay_factor=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = TimeDecay(decay_factor=-1)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = TimeDecay(decay_factor=2)
            lrs.on_train_begin()       
    

    def test_time_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_time_decay.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = TimeDecay(eta0=0.1, eta_min=0.02,
                        decay_factor=0.5)
        observers = [lrs]              
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg

@mark.observer
@mark.lrs
@mark.sqrt_time_decay
class SqrtTimeDecayTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_sqrt_time_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = SqrtTimeDecay(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = SqrtTimeDecay(eta0=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = SqrtTimeDecay(eta0=1)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = SqrtTimeDecay(eta_min=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = SqrtTimeDecay(eta_min=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = SqrtTimeDecay(eta_min=1)
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = SqrtTimeDecay(decay_factor=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = SqrtTimeDecay(decay_factor=-1)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = SqrtTimeDecay(decay_factor=2)
            lrs.on_train_begin()       
    

    def test_sqrt_time_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_sqrt_time_decay.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = SqrtTimeDecay(eta0=0.1, eta_min=0.045,
                        decay_factor=0.5)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg    

@mark.observer
@mark.lrs
@mark.exponential_decay
class ExponentialDecayTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_exponential_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = ExponentialDecay(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialDecay(eta0=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialDecay(eta0=1)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = ExponentialDecay(eta_min=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialDecay(eta_min=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialDecay(eta_min=1)
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = ExponentialDecay(decay_factor=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialDecay(decay_factor=-1)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialDecay(decay_factor=2)
            lrs.on_train_begin()       
    
    def test_exponential_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_exponential_decay.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = ExponentialDecay(eta0=0.1, eta_min=0.05,
                        decay_factor=0.1)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg                

@mark.observer
@mark.lrs
@mark.exponential_step_decay
class ExponentialStepDecayTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_exponential_step_decay_validation(self, get_mock_estimator):
        est = get_mock_estimator
        # Validate initial learning rate
        with pytest.raises(TypeError):            
            lrs = ExponentialStepDecay(eta0='h')
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta0=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta0=1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = ExponentialStepDecay(eta_min=None)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta_min=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta_min=1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = ExponentialStepDecay(decay_factor=None)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(decay_factor=-1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(decay_factor=2)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()       
        # Validate decay_steps
        with pytest.raises(TypeError):
            lrs = ExponentialStepDecay(decay_steps=None)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(decay_steps=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin() 
    
    def test_exponential_step_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_exponential_step_decay.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = ExponentialStepDecay(eta0=0.1, eta_min=0.01,
                        decay_factor=0.5, decay_steps=5)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg          

@mark.observer
@mark.lrs
@mark.exponential_step_decay_staircase
class ExponentialStepDecayStaircaseTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_exponential_step_decay_staircase_validation(self, get_mock_estimator):
        est = get_mock_estimator
        # Validate initial learning rate
        with pytest.raises(TypeError):            
            lrs = ExponentialStepDecay(eta0='h')
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta0=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta0=1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = ExponentialStepDecay(eta_min=None)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta_min=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(eta_min=1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = ExponentialStepDecay(decay_factor=None)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(decay_factor=-1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(decay_factor=2)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()       
        # Validate decay_staircase_steps
        with pytest.raises(TypeError):
            lrs = ExponentialStepDecay(decay_steps=None)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = ExponentialStepDecay(decay_steps=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin() 
    
    def test_exponential_step_decay_staircase(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_exponential_step_decay_staircase.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = ExponentialStepDecay(eta0=0.1, eta_min=0.02,
                        decay_factor=0.5, decay_steps=5, staircase=True)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg          

@mark.observer
@mark.lrs
@mark.polynomial_decay
class PolynomialDecayTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_polynomial_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = PolynomialDecay(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialDecay(eta0=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialDecay(eta0=1)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = PolynomialDecay(eta_min=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialDecay(eta_min=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialDecay(eta_min=1)
            lrs.on_train_begin()      
        # Validate power
        with pytest.raises(TypeError):
            lrs = PolynomialDecay(power=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialDecay(power=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialDecay(power=2)
            lrs.on_train_begin()       
    
    def test_polynomial_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_polynomial_decay.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = PolynomialDecay(eta0=0.1, eta_min=0.025,
                        power=0.95)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg           

@mark.observer
@mark.lrs
@mark.polynomial_step_decay
class PolynomialStepDecayTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_polynomial_step_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = PolynomialStepDecay(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialStepDecay(eta0=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialStepDecay(eta0=1)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = PolynomialStepDecay(eta_min=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialStepDecay(eta_min=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialStepDecay(eta_min=1)
            lrs.on_train_begin()      
        # Validate power
        with pytest.raises(TypeError):
            lrs = PolynomialStepDecay(power=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialStepDecay(power=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialStepDecay(power=2)
            lrs.on_train_begin()   
        # Validate decay_steps
        with pytest.raises(TypeError):
            lrs = PolynomialStepDecay(decay_steps=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PolynomialStepDecay(decay_steps=0)
            lrs.on_train_begin()                 
    
    def test_polynomial_step_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_polynomial_step_decay.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = PolynomialStepDecay(eta0=0.1, eta_min=0.001,
                        power=0.5, decay_steps=5)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg         

@mark.observer
@mark.lrs
@mark.power_decay
class PowerScheduleTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_power_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = PowerSchedule(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PowerSchedule(eta0=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PowerSchedule(eta0=1)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = PowerSchedule(eta_min=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PowerSchedule(eta_min=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PowerSchedule(eta_min=1)
            lrs.on_train_begin()      
        # Validate power
        with pytest.raises(TypeError):
            lrs = PowerSchedule(power=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PowerSchedule(power=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PowerSchedule(power=2)
            lrs.on_train_begin()       
        # Validate decay_steps
        with pytest.raises(TypeError):
            lrs = PowerSchedule(decay_steps=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = PowerSchedule(decay_steps=0)
            lrs.on_train_begin()                
    
    def test_power_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_power.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = PowerSchedule(eta0=0.1, eta_min=0.04,
                        power=0.95, decay_steps=5)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg       

@mark.observer
@mark.lrs
@mark.bottou_decay
class BottouScheduleTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['lr'],
                                    usecols="B").to_numpy().flatten()

    def test_bottou_decay_validation(self):
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = BottouSchedule(eta0='h')
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = BottouSchedule(eta0=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = BottouSchedule(eta0=1)
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = BottouSchedule(eta_min=None)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = BottouSchedule(eta_min=0)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = BottouSchedule(eta_min=1)
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = BottouSchedule(decay_factor=None)            
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = BottouSchedule(decay_factor=-1)
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = BottouSchedule(decay_factor=2)
            lrs.on_train_begin()       
            
    
    def test_bottou_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_bottou.xlsx")
        exp_results = self._get_expected_results(filepath)
        # Instantiate learning rate schedule and create it as an observer
        lrs = BottouSchedule(eta0=0.1, eta_min=0.07,
                        decay_factor=0.5)
        observers=[lrs]
        # Instantiate and fit mock estimator
        est = Estimator(epochs=10, eta0=0.1, observers=observers)
        est.fit()
        # Extract learning rate history
        epochs = est.blackbox_.epoch_log.get('epoch')
        act_results = est.blackbox_.epoch_log.get('eta')
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg                

@mark.observer
@mark.lrs
@mark.improvement
class ImprovementTests:

    def _get_expected_results(self, filepath):
        return pd.read_excel(filepath, sheet_name='results', header=0, names=['cost','lr'],
                                    usecols="C,H")

    def test_improvement_decay_validation(self, get_mock_estimator):
        est = get_mock_estimator        
        # Validate initial learning rate
        with pytest.raises(TypeError):
            lrs = Improvement(eta0='h')
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = Improvement(eta0=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = Improvement(eta0=1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()            
        # Validate minimum learning rate
        with pytest.raises(TypeError):
            lrs = Improvement(eta_min=None)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = Improvement(eta_min=0)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = Improvement(eta_min=1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()      
        # Validate decay_factor
        with pytest.raises(TypeError):
            lrs = Improvement(decay_factor=None)            
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = Improvement(decay_factor=-1)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()
        with pytest.raises(ValueError):
            lrs = Improvement(decay_factor=2)
            lrs.set_model(est(epochs=10))
            lrs.on_train_begin()   
    
    def test_improvement_decay(self, get_mock_estimator):       
        Estimator = get_mock_estimator
        # Obtain expected results
        filepath = os.path.join(testdatadir, "test_learning_rate_schedules_improvement.xlsx")
        exp_results = self._get_expected_results(filepath)
        cost = exp_results['cost'].values
        exp_results = exp_results['lr'].values
        # Instantiate learning rate schedule and create it as an observer
        lrs = Improvement(eta0=0.1, eta_min=0.007,
                        decay_factor=0.5, epsilon=0.01, patience=2)
        observers=[lrs]    
        # Instantiate mock estimator
        est = Estimator(epochs=20, eta0=0.1, observers=observers)       
        lrs.set_model(est)
        # Simulate training 
        lrs.on_train_begin()
        act_results = []
        lr = 0.1
        for i in range(20):            
            log = {'train_cost':cost[i], 'eta': lrs.model.eta}    
            lrs.on_epoch_end(epoch=i, log=log)                
            act_results.append(lrs.model.eta)
        # Compare two arrays
        act_res_len = len(act_results)
        exp_res_len = len(exp_results)
        msg = "Expected results length = {e}, actual results length = {a}".format(e=str(exp_res_len),
                                                                                  a=str(act_res_len))
        assert act_res_len == exp_res_len, msg
        msg = "Expected results {e}\nActual Results {a}".format(e=str(exp_results),a=str(act_results))
        assert np.allclose(exp_results, act_results), msg         


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# ============================================================================ #
# Project : MLStudio                                                           #
# Version : 0.1.0                                                              #
# File    : observers.py                                                       #
# Python  : 3.8.2                                                              #
# ---------------------------------------------------------------------------- #
# Author  : John James                                                         #
# Company : DecisionScients                                                    #
# Email   : jjames@decisionscients.com                                         #
# URL     : https://github.com/decisionscients/MLStudio                        #
# ---------------------------------------------------------------------------- #
# Created       : Sunday, March 15th 2020, 7:27:16 pm                          #
# Last Modified : Sunday, March 15th 2020, 7:37:00 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                      #
# ---------------------------------------------------------------------------- #
# License : BSD                                                                #
# Copyright (c) 2020 DecisionScients                                           #
# ============================================================================ #
"""Module containing functionality called during the training process.

Note: The ObserverList and Observer abstract base classes were inspired by
the Keras implementation.  
"""
from abc import ABC, abstractmethod, ABCMeta
import warnings
warnings.filterwarnings("once", category=RuntimeWarning, module='base')

import datetime
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator

from mlstudio.utils.validation import validate_int, validate_zero_to_one
from mlstudio.utils.validation import validate_metric
# --------------------------------------------------------------------------- #
#                             CALLBACK LIST                                   #
# --------------------------------------------------------------------------- #
class ObserverList:
    """Container of observers."""

    def __init__(self, observers=None):
        """ObserverList constructor
        
        Parameters
        ----------
        observers : list
            List of 'Observer' instances.        
        """
        observers = observers or []
        self.observers = [c for c in observers]        
        self.params = {}
        self.model = None

    def append(self, observer):
        """Appends observer to list of observers.
        
        Parameters
        ----------
        observer : Observer instance            
        """
        self.observers.append(observer)

    def set_params(self, params):
        """Sets the parameters variable, and in list of observers.
        
        Parameters
        ----------
        params : dict
            Dictionary containing model parameters
        """
        self.params = params
        for observer in self.observers:
            observer.set_params(params)

    def set_model(self, model):
        """Sets the model variable, and in the list of observers.
        
        Parameters
        ----------
        model : Estimator or subclass instance 
        
        """
        self.model = model
        for observer in self.observers:
            observer.set_model(model)

    def on_batch_begin(self, batch, log=None):
        """Calls the `on_batch_begin` methods of its observers.

        Parameters
        ----------
        batch : int
            Current training batch

        log: dict
            Currently no data is set to this parameter for this class. This may
            change in the future.
        """
        log = log or {}
        for observer in self.observers:
            observer.on_batch_begin(batch, log)

    def on_batch_end(self, batch, log=None):
        """Calls the `on_batch_end` methods of its observers.
        
        Parameters
        ----------
        batch : int
            Current training batch
        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """
        log = log or {}
        for observer in self.observers:
            observer.on_batch_end(batch, log)

    def on_epoch_begin(self, epoch, log=None):
        """Calls the `on_epoch_begin` methods of its observers.

        Parameters
        ----------        
        epoch: integer
            Current training epoch

        log: dict
            Currently no data is passed to this argument for this method
            but that may change in the future.
        """
        log = log or {}
        for observer in self.observers:
            observer.on_epoch_begin(epoch, log)

    def on_epoch_end(self, epoch, log=None):
        """Calls the `on_epoch_end` methods of its observers.
        This function should only be called during train mode.

        Parameters
        ----------
        epoch: int
            Current training epoch
        
        log: dict
            Metric results for this training epoch, and for the
            validation epoch if validation is performed.
        """
        log = log or {}
        for observer in self.observers:
            observer.on_epoch_end(epoch, log)

    def on_train_begin(self, log=None):
        """Calls the `on_train_begin` methods of its observers.

        Parameters
        ----------
        log: dict
            Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        for observer in self.observers:
            observer.on_train_begin(log)

    def on_train_end(self, log=None):
        """Calls the `on_train_end` methods of its observers.

        Parameters
        ----------
        log: dict
            Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        for observer in self.observers:
            observer.on_train_end(log)

    def __iter__(self):
        return iter(self.observers)

# --------------------------------------------------------------------------- #
#                             CALLBACK CLASS                                  #
# --------------------------------------------------------------------------- #
class Observer(ABC, BaseEstimator):
    """Abstract base class used to build new observers."""
    def __init__(self):
        """Observer class constructor."""
        self.params = None
        self.model = None

    def set_params(self, params):
        """Sets parameters from estimator.

        Parameters
        ----------
        params : dict
            Dictionary containing estimator parameters
        """ 
        self.params = params

    def set_model(self, model):
        """Stores model in Observer object.

        Parameters
        ----------
        model : Estimator
            Estimator object
        """
        self.model = model

    def on_batch_begin(self, batch, log=None):
        """Logic executed at the beginning of each batch.

        Parameters
        ----------
        batch : int
            Current training batch
        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """        
        pass

    def on_batch_end(self, batch, log=None):   
        """Logic executed at the end of each batch.
        
        Parameters
        ----------
        batch : int
            Current training batch
        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """                
        pass

    def on_epoch_begin(self, epoch, log=None):
        """Logic executed at the beginning of each epoch.
        
        Parameters
        ----------
        epoch : int
            Current epoch
        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """                
        pass

    def on_epoch_end(self, epoch, log=None):
        """Logic executed at the end of each epoch.
        
        Parameters
        ----------
        epoch : int
            Current epoch
        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """                      
        pass

    def on_train_begin(self, log=None):
        """Logic executed at the beginning of training.
        
        Parameters
        ----------        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """                      
        pass

    def on_train_end(self, log=None):
        """Logic executed at the end of training.
        
        Parameters
        ----------        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """               
        pass
# --------------------------------------------------------------------------- #
#                             PERFORMANCE BASE                                #
# --------------------------------------------------------------------------- #
class PerformanceObserver(Observer):
    """Base class for performance observers."""

    def __init__(self, metric='val_score', epsilon=1e-3, patience=5): 
        super(PerformanceObserver, self).__init__()       
        self.name = "Performance Base Observer"
        self.metric = metric                
        self.epsilon = epsilon
        self.patience = patience     

    @property
    def stabilized(self):
        return self._stabilized   
       
    def _validate(self):     
        if 'val' in self.metric and self.val_size == 0.0:   
            msg = "The val_size parameter must be greater than zero if the metric is {m}".format(m=self.metric)
            raise ValueError(msg)
        validate_zero_to_one(param=self.epsilon, param_name='epsilon',
                             left='closed', right='closed')       
        validate_int(param=self.patience, param_name='patience',
                     minimum=0, left='open', right='open')

    def on_train_begin(self, log=None):                
        """Sets key variables at beginning of training.        
        
        Parameters
        ----------
        log : dict
            Contains no information
        """
        super(PerformanceObserver, self).on_train_begin(log=log)        
        log = log or {}        
        self._validate()
        # Private variables
        self._baseline = None        
        # If the metric is not available in the log, we fallback to 'train_cost'        
        self._fallback_metric = 'train_cost'
        self._iter_no_improvement = 0
        self._stabilized = False
        self._significant_improvement = False
        self._critical_points = []

        # If the metric is a score, determine what constitutes a better
        # and best score. This can be obtained from the scorer object on 
        # the model. 
        if 'score' in self.metric:
            try:
                self._best = self.model.scorer.best
                self._better = self.model.scorer.better
        # If no scorer object on the estimator, then we will assume a greater 
        # score is better and the maximum score is best. This is consistent 
        # with default scorers for regression (R2) and classification (accuracy).                 
            except:
                self._best = np.max
                self._better = np.greater
        # Otherwise, the metric is cost and best and better costs are min and
        # less, respectively
        else:
            self._best = np.min            
            self._better = np.less            

        # Validation
        validate_metric(self.metric)
        validate_zero_to_one(param=self.epsilon, param_name='epsilon',
                             left='open', right='open')
        validate_int(param=self.patience, param_name='patience')

        # log data
        self.performance_log_ = {}

    def _update_log(self, epoch, log):
        """Creates log dictionary of lists of performance results."""
        log['epoch'] = epoch
        log['baseline']= self._baseline
        log['relative_change'] = self._relative_change
        log['significant_improvement'] = self._significant_improvement
        log['iter_no_improvement']= self._iter_no_improvement
        log['stabilized'] = self._stabilized
        for k,v in log.items():
            self.performance_log_.setdefault(k,[]).append(v)
        return log

    def _metric_improved(self, current):
        """Returns true if the direction and magnitude of change indicates improvement"""        
        return self._better(current, self._baseline)

    def _significant_relative_change(self, current):   
        """Returns true if relative change is greater than epsilon."""     
        self._relative_change = abs(current-self._baseline) / abs(self._baseline)
        return self._relative_change > self.epsilon                

    def _process_improvement(self, current, log=None):
        """Sets values of parameters and attributes if improved."""
        self._iter_no_improvement = 0            
        self._stabilized = False
        self._baseline = current
        self._best_results = log

    def _process_no_improvement(self, current, log=None):
        """Sets values of parameters and attributes if no improved."""    
        self._iter_no_improvement += 1  
        if self._iter_no_improvement == self.patience:
            self._stabilized = True
            # We reset iter_no_improvement and baseline to better of the 
            # current value and prior baseline. This gives the estimator
            # another 'patience' epochs to achieve real improvement from
            # new baseline.  
            self._iter_no_improvement = 0
            self._baseline = self._best((current, self._baseline)) 
        else:
            self._stabilized = False               

    def _get_current_value(self, log):
        """Obtain the designated metric or fallback metric from the log."""
        current = log.get(self.metric)
        if current is None:
            if self.metric != self._fallback_metric:
                current = log.get(self._fallback_metric)
                if current is None:
                    msg = "Metrics {m} and {fbm} are not available in the log.\
                        Check dimensions of input data to confirm that there\
                            is adequate data for the metric.".format(
                                m=self.metric, fbm=self._fallback_metric
                            )
                    raise Exception(msg)
                else:
                    msg = "Metric {m} was not available in the log. Using {fbm}\
                        metric instead.".format(m=self.metric, fbm=self._fallback_metric)
                    warnings.warn(message=msg, category=RuntimeWarning)
            else:
                msg = "Metric {m} was not available in the log. Check dimensions\
                    of input data to ensure that {m} can be computed.".format(m=self.metric)
                raise Exception(msg)    
            
        return current

    def on_epoch_end(self, epoch, log=None):
        """Logic executed at the end of each epoch.
        
        Parameters
        ----------
        epoch : int
            Current epoch
        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """                  
        log = log or {}   
        
        # Initialize state variables        
        self._significant_improvement = False
        self._relative_change = 0
        
        # Obtain current performance
        current = self._get_current_value(log)

        # Handle first iteration as an improvement by default
        if self._baseline is None:                            
            self._significant_improvement = True
            self._process_improvement(current, log)    

        # Otherwise, evaluate the direction and magnitude of the change        
        else:
            self._significant_improvement = self._metric_improved(current) and \
                self._significant_relative_change(current)

            if self._significant_improvement:
                self._process_improvement(current, log)
            else:
                self._process_no_improvement(current, log)

        # Log results and critical points
        log = self._update_log(epoch, log)
        self._critical_points.append(log)



        



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_objectives.py                                                #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, June 15th 2020, 3:45:31 pm                          #
# Last Modified : Monday, June 15th 2020, 3:45:31 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
#%%
import math
import os
from pathlib import Path
import sys

import glob
import numpy as np
import pandas as pd
import pytest
from pytest import mark
from scipy.special import softmax
from sklearn.metrics import mean_squared_error
from sklearn.datasets import make_regression, make_classification
from sklearn.datasets import make_multilabel_classification
from sklearn.preprocessing import LabelBinarizer

homedir = str(Path(__file__).parents[2])
datadir = os.path.join(homedir, "tests\\test_data")
sys.path.append(homedir)
sys.path.append(datadir)

from mlstudio.utils.data_manager import StandardScaler
from mlstudio.supervised.core.objectives import MSE, CrossEntropy
from mlstudio.supervised.core.objectives import CategoricalCrossEntropy
from mlstudio.supervised.core.regularizers import L1, L2, L1_L2

# --------------------------------------------------------------------------  #
def create_regression_prediction(y=None, X=None, theta=None):
    """Creates vector of predictions based upon target plus random noise."""
    noise = np.random.normal(0,1, size=y.shape)    
    y_pred = np.add(y,noise)
    return y_pred

def create_classification_prediction(y=None, X=None, theta=None):
    """Creates classification prediction as probability [0,1]"""
    return np.random.uniform(0,1, size=y.shape)

def create_multiclass_prediction(y=None, X=None, theta=None):
    """Creates multiclassification prediction."""
    z = X.dot(theta)
    return softmax(z, axis=1)

def make_regression_data():
    X, y = make_regression(n_samples=100, n_features=5, random_state=5)    
    scaler = StandardScaler()    
    X = scaler.fit_transform(X)    
    return X, y

def make_classification_data():
    X, y, = make_classification(n_samples=100, n_features=5, random_state=5)        
    scaler = StandardScaler()    
    X = scaler.fit_transform(X)    
    return X, y

def make_multiclass_data():
    X, y, = make_classification(n_samples=100, n_features=5, n_classes=4, 
                                n_informative=3, random_state=5)     
    enc = LabelBinarizer()
    y = enc.fit_transform(y) 
    scaler = StandardScaler()    
    X = scaler.fit_transform(X)    
    print(X.shape)
    print(y.shape)
    return X, y            

def create_data():
    # Designate filenames and create filepaths
    mse_filename = "test_objective_cost_functions_mse.xlsx"
    xe_filename = "test_objective_cost_functions_xe.xlsx"
    cxe_filename = "test_objective_cost_functions_cxe.xlsx"    
    mse_filepath = os.path.join(datadir, mse_filename)
    xe_filepath = os.path.join(datadir, xe_filename)
    cxe_filepath = os.path.join(datadir, cxe_filename)
    # Obtain data
    X_reg, y_reg = make_regression_data()
    X_bin, y_bin = make_classification_data()
    X_multi, y_multi = make_multiclass_data()
    # Create parameters
    regression_theta = np.random.default_rng().uniform(low=0, high=1, size=X_reg.shape[1])            
    classification_theta = np.random.default_rng().uniform(low=0, high=1, size=X_bin.shape[1])            
    multiclass_theta = np.random.default_rng().uniform(low=0, high=1, size=(X_multi.shape[1],y_multi.shape[1]))            
    # Create packages 
    regression_pack = {'locked': True, 'filepath': mse_filepath, 'X':X_reg, 
                       'y': y_reg, 'theta': regression_theta, 
                       'predict': create_regression_prediction}
    classification_pack = {'locked': True, 'filepath': xe_filepath, 
                           'X':X_bin, 'y': y_bin, 'theta': classification_theta,
                           'predict': create_classification_prediction}
    multiclass_pack = {'locked': True, 'filepath': cxe_filepath, 'X':X_multi, 
                       'y': y_multi, 'theta': multiclass_theta,
                       'predict': create_multiclass_prediction}
    data_packs = [regression_pack, classification_pack, multiclass_pack]
    # Write to files
    for data in data_packs:        
        if data['locked']:
            pass
        else:
            X = pd.DataFrame(data=data['X'])
            y = pd.DataFrame(data=data['y'])        
            theta = pd.DataFrame(data=data['theta'])
            y_pred = pd.DataFrame(data=data['predict'](y, X, theta))
            with pd.ExcelWriter(data['filepath']) as writer:
                X.to_excel(writer, sheet_name='X')
                y.to_excel(writer, sheet_name='y')
                y_pred.to_excel(writer, sheet_name='y_pred')
                theta.to_excel(writer, sheet_name='theta')

#create_data()
#%%


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : optimization.py                                                   #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, May 25th 2020, 3:59:09 am                           #
# Last Modified : Monday, May 25th 2020, 3:59:09 am                           #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Functions used to diagnose optimization of a model."""
import numpy as np
import pandas as pd
from tabulate import tabulate

from mlstudio.utils.data_analyzer import describe_numeric_array

def diagnose_gradient(estimator):
    """Produces descriptive statistics for the gradient."""
    gradient_norms = estimator.blackbox_.epoch_log.get('gradient_norm')    
    df = describe_numeric_array(gradient_norms, fmt='df')            
    print(tabulate(df, headers='keys', tablefmt='psql', showindex=False))
    return df





#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : model_validation.py                                               #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, March 20th 2020, 10:16:08 pm                        #
# Last Modified : Friday, March 20th 2020, 10:16:08 pm                        #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Functions used to analyze model validity."""

import math
from math import erf
import numpy as np
from scipy.stats import norm

# --------------------------------------------------------------------------- #
#                           RESIDUAL ANALYSIS                                 #
# --------------------------------------------------------------------------- #
def standardized_residuals(X, y, y_pred):
    """Computes standardized residuals.
    
    Standardized residuals (sometimes referred to as "internally studentized 
    residuals") are defined for each observation, i = 1, ..., n as an 
    ordinary residual divided by an estimate of its standard deviation:
    ..math:: r_i = \\frac{e_i}{\\sqrt{MSE(1-h_{ii})}}

    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features
    
    y : ndarray or Series of length n
        An array or series of target or class values 

    y_pred : ndarray or Series of length n
        An array or series of predictions         
    """
    # Compute residuals
    residuals = y - y_pred
    
    # Compute Leverage
    hii = leverage(X)
    
    # Compute degrees of freedom and MSE
    rank = np.linalg.matrix_rank(X)
    df = X.shape[0] - rank
    mse = np.matmul(residuals, residuals) / df

    # Calculate standardized 
    standardized_residuals = residuals / np.sqrt(mse  * (1-hii))
    
    return standardized_residuals

def studentized_residuals(X, y, y_pred):
    """Computes studentized residuals.

    Studentized residuals are just a deleted residual divided by its estimated
    standard deviation. This turns out to be equivalent to the ordinary residual
    divided by a factor that includes the mean square error based on the 
    estimated model with the ith observation deleted, MSE(i), and the leverage, hii
    .. math:: r_i = \\frac{e_i}{\\sqrt{MSE_{(i)}(1-h_{ii})}}
    
    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features
    
    y : ndarray or Series of length n
        An array or series of target or class values 

    y_pred : ndarray or Series of length n
        An array or series of predictions         
    """    
    # Using the calculation from 
    # https://newonlinecourses.science.psu.edu/stat462/node/247/
    n = X.shape[0]
    k = X.shape[1]
    r = standardized_residuals(X=X, y=y, y_pred=y_pred)

    # Calculate studentized residuals 
    studentized_residuals = r * np.sqrt((n-k-2)/(n-k-1-np.square(r)))

    return studentized_residuals   

# --------------------------------------------------------------------------- #
#                               QUANTILES                                     #
# --------------------------------------------------------------------------- #
def quantile(p):
    """Computes the theoretical quantiles for random variable.

    Parameters
    ----------
    p : array-like of real-valued numbers
        The vector of real numbers for which the theoretical quantiles 
        are computed.


    """
    j = np.arange(len(p)) + 1
    j_star = (j-0.5) / len(p)
    return norm.ppf(j_star)    
     
def rational_approximation(t):

    # Abramowitz and Stegun formula 26.2.23.
    c = [2.515517, 0.802853, 0.010328]
    d = [1.432788, 0.189269, 0.001308]
    numerator = (c[2]*t + c[1])*t + c[0]
    denominator = ((d[2]*t + d[1])*t + d[0])*t + 1.0
    return t - numerator / denominator
    

def normal_CDF_inverse(p):

    assert p > 0.0 and p < 1

    # See article above for explanation of this section.
    if p < 0.5:
        # F^-1(p) = - G^-1(p)
        return -rational_approximation( math.sqrt(-2.0*math.log(p)) )
    else:
        # F^-1(p) = G^-1(1-p)
        return rational_approximation( math.sqrt(-2.0*math.log(1.0-p)) )   
# --------------------------------------------------------------------------- #
#                           INFLUENCE MEASURES                                #
# --------------------------------------------------------------------------- #
def leverage(X):
    """Computes leverage scores for a data set.

    Leverage is a measure of how much the independent variables of an 
    observation differs from the other observations. The leverage score for
    the ith osbservation is defined as:
    .. math:: h_{ii} = 
    the ith diagonal element of the project matrix \\mathbb{H}=X(X^TX)^{-1}X^T,
    where X is the design matrix whose rows are observations and columns
    are the independent variables.
    
    The diagonal of the projection matrix, commonly known as the hat matrix, 
    is a standardized measure of the distance from the ith observation from
    the center (or centroid) of the x-space.
    
    Points with leverage greater than \\frac{2p}{n}, where p is the number
    independent variables, including the bias, and n is the number of 
    observations, are considered remote enough from the rest of the 
    data to be designated a leverage point. [2]_
    
    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features
    
    Returns
    -------
    leverage : ndarray or DataFrame of shape n x 1
        Contains the leverage scores for each observation.
    """

    hat = X.dot(np.linalg.inv(X.T.dot(X)).dot(X.T))
    hii = np.diagonal(hat)
    return hii

def cooks_distance(X, y, y_pred):
    """Computes Cook's Distance, a commonly used measure of data point influence.
    
    Cook's Distance is used in least-squares regression analysis to identify 
    influencial data points. Cook's distance for a given data point is given by:
    .. math:: \\mathbb{D}_i = \\frac{\\epsilon^2_i}{ps^2}\\big[\\frac{h_{ii}}{(1-h_{ii})^2}][1]_
    
    An alternative formulation relates Cooks Distance to studentized
    residuals. 
    .. math:: \\mathbb{D}_i =\\bigg[\\frac{1}{p}\\frac{n}{n-p}]t_i^2\\frac{h_{ii}}{1-h{ii}}
    
    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features
    
    y : ndarray or Series of length n
        An array or series of target or class values 

    y_pred : ndarray or Series of length n
        An array or series of predictions         
    
    Returns
    -------
    cooks_distance : ndarray or DataFrame of shape n x 1
        A matrix of cooks distances computed for each observation
    
    Reference
    ---------
    .. [1]  Wikipedia contributors. (2019, October 24). Cook's distance. 
            In Wikipedia, The Free Encyclopedia. Retrieved 01:32, 
            November 29, 2019, from 
            https://en.wikipedia.org/w/index.php?title=Cook%27s_distance&oldid=922838890
    """
    # Compute residual
    e = y - y_pred
    # Set number of observations and predictors
    n = X.shape[0]
    p = X.shape[1]
    # S squared 
    s2 = np.matmul(e.T,e) / (n-p)
    # Compute leverage
    hii = leverage(X)
    # Obtain studentized residuals 
    cooks_d = (e**2/ (p * s2)) * (hii/(1-hii)**2)    
    return cooks_d

def dffits(X, y, y_pred):
    """Computes the difference in fits (DFFITS).

    The difference in fits for observation i, denoted (\\DFFITS_i), is defined
    as:
    .. math:: \\DFFITS_i = \\frac{\\hat{y}_i-\\hat{y}_{(i)}}{\\sqrt{MSE_{(i)}h_{ii}}}.
    
    The numerator measures the difference in the predicted response obtained
    with and without the ith data point. The denominator is the estimated 
    standard deviation of the difference between predicted responses. Hence,
    the diffeence in fits quantifies the number of standard deviations that 
    the fitted value changes when the ith observation is omitted.
    
    An observation is deemed influential if the absolute value of its DFFITS
    value is greater than:
    .. math:: 2\\sqrt{\\frac{p+1}{n-p-1}}
    
    where n is the number of observations, p is the number of predictors 
    including the bias term.
    
    DFFITS is also related to the students residual, and is in fact the latter
    times 
    
    .. math:: \\sqrt(\\frac{h_{ii}}{1-h_{ii}})[1]_
    
    Observations with DFFITS greater that 2\\sqrt{\\frac{p}{n}}, where p is the 
    number of predictors including the bias and n is the number of observations,
    should be investigated.[2]_
    
    Parameters
    ----------
    model : Estimator or BaseEstimator
        ML Studio or Scikit Learn estimator
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features
    y : ndarray or Series of length n
        An array or series of target or class values 
    
    References
    ----------    
    .. [1] Montogomery, Douglas C.; Peck, Elizabeth A.; Vining, G. 
           Geoffrey (2012). Introduction to Linear Regression Analysis 
           (5th ed.). Wiley. p. 218. ISBN 978-0-470-54281-1. Retrieved 
           22 February 2013. Thus, DFFITSi is the value of R-student 
           multiplied by the leverage of the ith observation [hii/(1-hii)]1/2.
    .. [2] Belsley, David A.; Kuh, Edwin; Welsh, Roy E. (1980). 
           Regression Diagnostics: Identifying Influential Data and 
           Sources of Collinearity. Wiley Series in Probability 
           and Mathematical Statistics. New York: John Wiley & Sons. 
           pp. 11–16. ISBN 0-471-05856-4.
    """
    r_student = studentized_residuals(X, y, y_pred)
    hii = leverage(X)
    df = r_student * np.sqrt(hii/(1-hii))
    return df


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : Data Studio                                                       #
# Version : 0.1.0                                                             #
# File    : print.py                                                          #
# Python  : 3.8.1                                                             #
# --------------------------------------------------------------------------- #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/datastudio                     #
# --------------------------------------------------------------------------- #
# Created       : Saturday, February 22nd 2020, 8:01:28 pm                    #
# Last Modified : Saturday, February 22nd 2020, 8:01:29 pm                    #
# Modified By   : John James (jjames@decisionscients.com>)                    #
# --------------------------------------------------------------------------- #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
""" Print utilities."""
from abc import ABC, abstractmethod
import math
import statistics
# --------------------------------------------------------------------------- #
#                                Print                                        #
# --------------------------------------------------------------------------- #
class Printer(ABC):
    """Abstract base class defining behaviors and interface for print classes."""

    def __init__(self, line_length = 80, title_separator='='):
        self._line_length = min(line_length,80)
        self._default_anchor_pos = int(math.floor(self._line_length / 2))
        self._default_anchor_style = ":"
        self._default_anchor_lhs_pad = 1
        self._default_anchor_rhs_pad = 1     
        self._title_separator = title_separator   

    def _compute_budgets(self, anchor):
        """Computes available space on left and right sides."""
        anchor['lhs_budget'] = anchor['pos'] - anchor['lhs_pad'] 
        anchor['rhs_budget'] = self._line_length - anchor['rhs_pad'] -\
             anchor['pos'] 
        return anchor
    
    def _set_anchor(self, content):
        """Sets vertical anchor point for text alignment."""

        # Initialize anchor, padding and budget with default values
        anchor = {}
        anchor['style'] = self._default_anchor_style
        anchor['pos'] = self._default_anchor_pos
        anchor['lhs_pad'] = self._default_anchor_lhs_pad
        anchor['rhs_pad'] = self._default_anchor_rhs_pad  
        # Compute budgets given default anchor position
        anchor = self._compute_budgets(anchor)    

        # Get line lengths from content
        lhs_lens = [len(k) for k in content.keys()]
        rhs_lens = [len(v) for v in content.values()]    

        # If all lengths are within budget return self._anchor_pos
        if max(lhs_lens) <= anchor['lhs_budget'] and \
           max(rhs_lens) <= anchor['rhs_budget']:
           return anchor

        # Otherwise adjust anchor by 1/2 (avg_rhs - avg_lhs)
        lhs_avg_len = statistics.mean(lhs_lens)
        rhs_avg_len = statistics.mean(rhs_lens)
        anchor['pos'] -= 1/2 * (rhs_avg_len - lhs_avg_len)
        # Adjust the budgets for lhs and rhs accordingly
        anchor = self._compute_budgets(anchor)

        return anchor    


    def print_title(self, title):
        """Prints title and title separator"""
        if not isinstance(self._line_length, (int, float)):
            raise TypeError("Invalid line length. Must be an integer or float.")
        title_separator = self._title_separator * len(title)
        print("\n")
        print(title.center(self._line_length))
        print(title_separator.center(self._line_length))

    def _print_line(self, anchor, k, v):
        lhs_pad = ' ' * int(anchor['pos'] - anchor['lhs_pad'] - len(k))
        line = lhs_pad + k + ' '*anchor['lhs_pad'] + anchor['style'] + \
            ' '*anchor['rhs_pad'] + v
        print(line)


    def print_dictionary(self, content, title=None):
        """Pretty prints a title and dictionary."""
        anchor = self._set_anchor(content) 
        if title:
            self.print_title(title)
        for k, v in content.items():
            self._print_line(anchor, k,v)

    def print_dataframe(self, content, title=None):
        """Prints a table with a heading.

        Note: To use the tabulate package, content must be a dictionary and 
        each key must be an iterable.

        Parameters
        ----------
        title : str
            The title to be printed above the table
        content : dict of lists
            Dictionary in which the values are iterables.
        """
        if title:
            self.print_title(title)
        print(content.to_string())

        



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : test_objectives.py                                                #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, June 15th 2020, 3:45:31 pm                          #
# Last Modified : Monday, June 15th 2020, 3:45:31 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
#%%
import math
import os
from pathlib import Path
import sys

import glob
import numpy as np
import pandas as pd
import pytest
from pytest import mark
from sklearn.metrics import mean_squared_error
from sklearn.datasets import make_regression, make_classification
from sklearn.datasets import make_multilabel_classification

homedir = str(Path(__file__).parents[3])
datadir = os.path.join(homedir, "tests\\test_data")
sys.path.append(homedir)
sys.path.append(datadir)

from mlstudio.supervised.core.objectives import MSE, CrossEntropy
from mlstudio.supervised.core.objectives import CategoricalCrossEntropy
from mlstudio.supervised.core.regularizers import L1, L2, L1_L2
# --------------------------------------------------------------------------  #
@mark.objectives
@mark.mse
class MSETests:

    @mark.mse_cost
    def test_cost(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE()
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost'][0])

    @mark.mse_cost
    def test_cost_l1(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE(L1(alpha=0.1))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l1'][0])

    @mark.mse_cost
    def test_cost_l2(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE(L2(alpha=0.1))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l2'][0])

    @mark.mse_cost        
    def test_cost_l1_l2(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE(L1_L2(alpha=0.1, ratio=0.5))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l1_l2'][0])    

    # ----------------------------------------------------------------------  #
    @mark.mse_grad
    def test_gradient(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE()
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad'])

    @mark.mse_grad
    def test_gradient_l1(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE(regularizer=L1(alpha=0.1))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l1'])

    @mark.mse_grad
    def test_gradient_l2(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE(regularizer=L2(alpha=0.1))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l2'])

    @mark.mse_grad
    def test_gradient_l1_l2(self, get_objective_mse_package):
        p = get_objective_mse_package
        obj = MSE(regularizer=L1_L2(alpha=0.1, ratio=0.5))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l1_l2'])
    

# --------------------------------------------------------------------------  #
@mark.objectives
@mark.cross_entropy
class CrossEntropyTests:

    @mark.cross_entropy_cost
    def test_cost(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy()
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost'][0])

    @mark.cross_entropy_cost
    def test_cost_l1(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy(L1(alpha=0.1))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l1'][0])

    @mark.cross_entropy_cost
    def test_cost_l2(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy(L2(alpha=0.1))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l2'][0])

    @mark.cross_entropy_cost        
    def test_cost_l1_l2(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy(L1_L2(alpha=0.1, ratio=0.5))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l1_l2'][0])    

    # ----------------------------------------------------------------------  #
    @mark.cross_entropy_grad
    def test_gradient(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy()
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad'])

    @mark.cross_entropy_grad
    def test_gradient_l1(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy(regularizer=L1(alpha=0.1))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l1'])

    @mark.cross_entropy_grad
    def test_gradient_l2(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy(regularizer=L2(alpha=0.1))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l2'])

    @mark.cross_entropy_grad
    def test_gradient_l1_l2(self, get_objective_xe_package):
        p = get_objective_xe_package
        obj = CrossEntropy(regularizer=L1_L2(alpha=0.1, ratio=0.5))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l1_l2'])
        
# --------------------------------------------------------------------------  #
@mark.objectives
@mark.multiclass
class CategoricalCrossEntropyTests:

    @mark.multiclass_cost
    def test_cost(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy()
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost'][0])

    @mark.multiclass_cost
    def test_cost_l1(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy(L1(alpha=0.1))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l1'][0])

    @mark.multiclass_cost
    def test_cost_l2(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy(L2(alpha=0.1))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l2'][0])

    @mark.multiclass_cost        
    def test_cost_l1_l2(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy(L1_L2(alpha=0.1, ratio=0.5))
        J = obj(p['theta'], p['y'], p['y_pred'])
        assert np.isclose(J, p['cost_l1_l2'][0])    

    # ----------------------------------------------------------------------  #
    @mark.multiclass_grad
    def test_gradient(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy()
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad'])

    @mark.multiclass_grad
    def test_gradient_l1(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy(regularizer=L1(alpha=0.1))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l1'])

    @mark.multiclass_grad
    def test_gradient_l2(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy(regularizer=L2(alpha=0.1))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l2'])

    @mark.multiclass_grad
    def test_gradient_l1_l2(self, get_objective_cxe_package):
        p = get_objective_cxe_package
        obj = CategoricalCrossEntropy(regularizer=L1_L2(alpha=0.1, ratio=0.5))
        g = obj.gradient(p['theta'], p['X'], p['y'], p['y_pred'])
        assert np.allclose(g, p['grad_l1_l2'])        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : data_analyzer.py                                                  #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Monday, March 23rd 2020, 9:17:20 am                         #
# Last Modified : Monday, March 23rd 2020, 9:17:21 am                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Data analysis helper functions."""
import numpy as np
import pandas as pd
from scipy.stats import skew, kurtosis
# --------------------------------------------------------------------------- #
def standardized_residuals(residuals):
    """Computes standardized residuals."""
    residuals = residuals.ravel()
    return residuals/np.std(residuals)  

# --------------------------------------------------------------------------- #
def uniform_order_stat(x):
    """Estimates uniform order statistics medians for the normal distribution."""
    positions = np.arange(1, len(x)+1)
    n = len(positions)
    u_i = (positions-0.375)/(n+0.25)
    return u_i
# --------------------------------------------------------------------------- #
def z_score(x):
    """Computes z-scores for a series of values."""
    mu = np.mean(x)
    std = np.std(x)
    z = (x-mu)/std
    return z

# --------------------------------------------------------------------------- #    
def theoretical_quantiles(x):
    """Computes the theoretical quantiles for a vector x."""
    u_i =  uniform_order_stat(x)
    q = z_score(u_i)
    return q

def sample_quantiles(x):
    """Computes the sample quantiles for a vector x."""
    x_sorted = np.sort(x)
    q = z_score(x_sorted)
    return q
    
# --------------------------------------------------------------------------  #
def cosine(a,b):
    """Returns the cosine similarity between two vectors."""
    numerator = a.dot(b)
    denominator = np.linalg.norm(a) * np.linalg.norm(b)
    return numerator / denominator

# --------------------------------------------------------------------------  #
def describe_numeric_array(x, fmt='dict'):
    """Returns descriptive statistics for a numeric array."""
    d = {}
    d['count'] = len(x)
    d['min'] = np.min(x)
    d['max'] = np.max(x)
    d['mean'] = np.mean(x)
    d['std'] = np.std(x)
    percentiles = [25, 50, 75]
    for p in percentiles:
        key = str(p) + 'th percentile' 
        d[key] = np.percentile(x, p)
    d['skew'] = skew(x, axis=None)
    d['kurtosis'] = kurtosis(x, axis=None)
    if fmt != 'dict':
        d = pd.DataFrame(d, index=[0])
    return d

def describe_categorical_array(x, fmt='dict'):
    """Returns descriptive statistics for a categorical array."""
    d = {}
    unique, pos = np.unique(x, return_inverse=True)
    counts = np.bincount(pos)
    maxpos = counts.argmax()
    d['count'] = len(x)
    d['unique'] = np.unique(x)
    d['top'] = unique[maxpos]
    d['freq'] = counts[maxpos]
    if fmt != 'dict':
        d = pd.DataFrame(d, index=[0])    
    return d






    



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : animations.py                                                     #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Wednesday, June 10th 2020, 8:34:00 pm                       #
# Last Modified : Wednesday, June 10th 2020, 8:44:50 pm                       #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Module for animations."""
from collections import OrderedDict
from copy import copy, deepcopy    
import datetime
import itertools
from itertools import zip_longest
import math
import os
from pathlib import Path
import sys

from IPython.display import HTML
import matplotlib as mpl
import matplotlib.pyplot as plt
plt.style.use('seaborn-whitegrid')
import matplotlib.animation as animation
from matplotlib.animation import FuncAnimation
from matplotlib import cm
from matplotlib import animation, rc, rcParams
rcParams['animation.embed_limit'] = 50
from matplotlib import colors as mcolors
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from mpl_toolkits.mplot3d import axes3d, Axes3D
import numpy as np
from numpy import array, newaxis
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go 
import plotly.io as pio 
import plotly.offline as py
from plotly.subplots import make_subplots
import seaborn as sns

from mlstudio.utils.data_manager import todf
from mlstudio.utils.file_manager import check_directory
from mlstudio.utils.file_manager import save_fig, save_csv, save_gif
# --------------------------------------------------------------------------  #
#                       ANIMATE OPTIMIZATION                                  #
# --------------------------------------------------------------------------  #
def animate_optimization(estimators, max_frames=None, filepath=None, show=True):
    """Renders surface plots and continuous lines to show optimizations.
    
    Parameters
    ----------
    estimators : dict
        Dictionary containing one or more MLStudio estimator objects.

    filepath : str (default=None)
        The path to which the animation is to be saved.

    show : bool (default=True)
        Indicates whether an animation should be rendered in users browser.

    """
    # ------------------------------------------------------------------  #
    # Extract parameter and cost data from the model blackboxes
    # Data for meshgrid
    xm, xM = 0, 0 # x range from objective function object
    ym, yM = 0, 0 # y range from objective function object
    # Models containing plot data (parameters theta and cost)
    models = OrderedDict()
    # Index for model dictionary  
    names = [] 
    # Data used to obtain best model for the objective function
    true = None
    estimates = []
    similarities = [] 
    # The number of frames = model.epochs
    n_frames = 0
    # The objective function
    objective = None
    
    for name, estimator in estimators.items():
        model = estimator['model']
        results = estimator['results']      
        # Get range from objective object for meshgrid
        x, y = model.objective.range
        xm, xM = x['min'], x['max']
        ym, yM = y['min'], y['max']
        # Extract model data for plotting gradient descent models         
        theta = model.blackbox_.epoch_log.get('theta')
        # Obtain step number for slicing if max_frames is provided.
        step = math.floor(len(theta) / max_frames) if max_frames else None
        # Clip thetas to plotting range 
        theta_0 = np.clip(np.array(theta['theta_0']), xm, xM)
        theta_1 = np.clip(np.array(theta['theta_1']), ym, yM)
        # Store individual thetas in dictionary 
        d = OrderedDict()
        d['theta_0'] = theta_0[::step]
        d['theta_1'] = theta_1[::step]
        d['cost'] = model.blackbox_.epoch_log.get('train_cost')[::step]
        models[name] = d
        # Place index for models dictionary in a list for easy access
        names.append(name)
        # Update best model data
        true = model.objective.minimum if true is None else true
        estimates.append(model.theta_)
        similarities.append(results['sim'])
        # Get objective function name for plot header 
        objective = model.objective if objective is None else objective
        # Designate the number of frames
        n_frames = len(d['theta_0']) if n_frames == 0 else n_frames

    # ------------------------------------------------------------------  #
    # Find best performing estimator and data
    best_idx = np.argmax(np.array(similarities))
    best_model = names[best_idx]
    best_true = str(true)
    best_est = str(np.around(estimates[best_idx],2))
    best_sim = str(np.around(similarities[best_idx],2))
    best_msg = "<b>Best Model: " + best_model + "</b><br>" + \
        "<b>True Minimum:</b> " + best_true +\
        "<b> Estimated Minimum:</b> " + best_est + \
            "<b> Cosine Similarity:</b> " + best_sim

    # ------------------------------------------------------------------  #
    # Create data for surface plot
    theta0_min, theta0_max = xm, xM
    theta1_min, theta1_max = ym, yM
    theta0_mesh = np.linspace(theta0_min, theta0_max, 50)
    theta1_mesh = np.linspace(theta1_min, theta1_max, 50)
    theta0_mesh_grid, theta1_mesh_grid = np.meshgrid(theta0_mesh, theta1_mesh)        
    # Create z axis grid based upon X,y and the grid of thetas
    Js = np.array([objective(THETA) for THETA in zip(np.ravel(theta0_mesh_grid), np.ravel(theta1_mesh_grid))])
    Js = Js.reshape(theta0_mesh_grid.shape)          

    # ------------------------------------------------------------------  #
    # Add colors to model
    colors = [ "red", "blue", "green", "purple", "orange", "black", "darkcyan", "maroon",
                "darkgoldenrod", "sienna", "darkslategrey", "lightcoral", "lime"]
    for i, model in enumerate(models.values()):
        model['color'] = colors[i]


    fig = go.Figure()
    # ------------------------------------------------------------------  #
    # Add Surface Plot        
    # Trace 0: Surface plot                
    fig.add_trace(
        go.Surface(x=theta0_mesh, y=theta1_mesh, z=Js, colorscale="RdBu", 
                    opacity=0.7, showscale=False, showlegend=False))

    # ------------------------------------------------------------------  #
    # Traces 1-12: Add Gradient Descent Trajectories
    for name, model in models.items():
        fig.add_trace(
            go.Scatter3d(x=model['theta_0'][:1], y=model['theta_1'][:1], z=model['cost'][:1],
                        name=name, 
                        showlegend=True, 
                        mode='lines', line=dict(color=model['color'], width=5)))                        

    # ------------------------------------------------------------------  #
    # Set layout title, font, template, etc...
    fig.update_layout(
        height=600, width=1200,
        scene=dict(
            xaxis=dict(nticks=20),
            zaxis=dict(nticks=4)
        ),
        #scene_xaxis=dict(range=[theta0_min, theta0_max], autorange=False),
        #scene_yaxis=dict(range=[theta1_min, theta1_max], autorange=False),            
        #scene_zaxis=dict(range=[zm, zM], autorange=False),            
        title=dict(text=objective.name, xanchor='center', yanchor='top', x=0.5, y=0.9),        
        font=dict(family="Open Sans"),                
        showlegend=True,      
        annotations=[
            dict(
                x=0.5,
                y=1.0,
                showarrow=False,
                text=best_msg,
                xref="paper",
                yref="paper"
            )               
        ],      
        template='plotly_white');                       

    # ------------------------------------------------------------------  #
    # Create frames                       
    frames = [go.Frame(
        dict(
            name = f'{k+1}',
            data = [                    
                go.Scatter3d(x=models[names[0]]['theta_0'][:k+1], 
                                y=models[names[0]]['theta_1'][:k+1], 
                                z=models[names[0]]['cost'][:k+1]),
                go.Scatter3d(x=models[names[1]]['theta_0'][:k+1], 
                                y=models[names[1]]['theta_1'][:k+1], 
                                z=models[names[1]]['cost'][:k+1]),                                 
                go.Scatter3d(x=models[names[2]]['theta_0'][:k+1], 
                                y=models[names[2]]['theta_1'][:k+1], 
                                z=models[names[2]]['cost'][:k+1]),
                go.Scatter3d(x=models[names[3]]['theta_0'][:k+1], 
                                y=models[names[3]]['theta_1'][:k+1], 
                                z=models[names[3]]['cost'][:k+1]),
                go.Scatter3d(x=models[names[4]]['theta_0'][:k+1], 
                                y=models[names[4]]['theta_1'][:k+1], 
                                z=models[names[4]]['cost'][:k+1]),                                 
                go.Scatter3d(x=models[names[5]]['theta_0'][:k+1], 
                                y=models[names[5]]['theta_1'][:k+1], 
                                z=models[names[5]]['cost'][:k+1]),
                go.Scatter3d(x=models[names[6]]['theta_0'][:k+1], 
                                y=models[names[6]]['theta_1'][:k+1], 
                                z=models[names[6]]['cost'][:k+1]),
                go.Scatter3d(x=models[names[7]]['theta_0'][:k+1], 
                                y=models[names[7]]['theta_1'][:k+1], 
                                z=models[names[7]]['cost'][:k+1]),   
                go.Scatter3d(x=models[names[8]]['theta_0'][:k+1], 
                                y=models[names[8]]['theta_1'][:k+1], 
                                z=models[names[8]]['cost'][:k+1]),   
                go.Scatter3d(x=models[names[9]]['theta_0'][:k+1], 
                                y=models[names[9]]['theta_1'][:k+1], 
                                z=models[names[9]]['cost'][:k+1]),   
                go.Scatter3d(x=models[names[10]]['theta_0'][:k+1], 
                                y=models[names[10]]['theta_1'][:k+1], 
                                z=models[names[10]]['cost'][:k+1]),   
                go.Scatter3d(x=models[names[11]]['theta_0'][:k+1], 
                                y=models[names[11]]['theta_1'][:k+1], 
                                z=models[names[11]]['cost'][:k+1]),                                                                                                                                                                                                                                                                                                                                         

            ],
            traces=[1,2,3,4,5,6,7,8,9,10,11,12])
        ) for k in range(n_frames-1)]

    # Update the menus
    updatemenus = [dict(type='buttons',
                        buttons=[dict(label="Play",
                                        method="animate",
                                        args=[[f'{k+1}' for k in range(n_frames-1)],
                                        dict(frame=dict(duration=1, redraw=True),
                                                transition=dict(duration=1),
                                                easing="linear",
                                                fromcurrent=True,
                                                mode="immediate")])],
                        direction="left",
                        pad=dict(r=10, t=85),
                        showactive=True, x=0.1, y=0, xanchor="right", yanchor="top")]

    sliders = [{"yanchor": "top",
                "xanchor": "left",
                "currentvalue": {"font": {"size": 16}, "prefix": "Iteration: ", "visible":True, "xanchor": "right"},
                'transition': {'duration': 1, 'easing': 'linear'},
                'pad': {'b': 10, 't': 50}, 
                'len': 0.9, 'x': 0.1, 'y': 0, 
                'steps': [{'args': [[f'{k+1}'], {'frame': {'duration': 1, 'easing': 'linear', 'redraw': False},
                                    'transition': {'duration': 1, 'easing': 'linear'}}], 
                    'label': k, 'method': 'animate'} for k in range(n_frames-1)       
                ]}]

    fig.update(frames=frames)

    fig.update_layout(
        updatemenus=updatemenus,
        sliders=sliders
    )

    if filepath:
        fig.write_html(filepath, include_plotlyjs='cdn', include_mathjax='cdn')
    if show:
        pio.renderers.default = "browser"
        fig.show()
# --------------------------------------------------------------------------  #
#                      ANIMATE OPTIMIZATION REGRESSION                        #
# --------------------------------------------------------------------------  #
def animate_optimization_regression(estimators, max_frames=None, filepath=None, show=True):
    """Renders subplots of the optimization and regression line fit.
    
    Parameters
    ----------
    estimators : dict
        Dictionary containing one or more MLStudio estimator objects.

    max_frames : int
        The maximum number of frames to show

    filepath : str (default=None)
        The path to which the animation is to be saved.

    show : bool (default=True)
        Indicates whether an animation should be rendered in users browser.

    """    
    # ------------------------------------------------------------------  #
    # Extract parameter and cost data from the model blackboxes
    theta0 = []
    theta1 = []
    models = OrderedDict()
    names = [] 
    for name, estimator in estimators.items():
        theta = estimator.blackbox_.epoch_log.get('theta')
        # Obtain step number for slicing if max_frames is provided.
        step = math.floor(len(theta) / max_frames) if max_frames else None
        # Thetas converted to individual columns in dataframe and extacted  
        theta = todf(theta, stub='theta_')
        theta0.extend(theta['theta_0'][::step].values.flatten())
        theta1.extend(theta['theta_1'][::step].values.flatten())
        d = OrderedDict()
        d['theta_0'] = theta['theta_0'][::step].values.flatten()
        d['theta_1'] = theta['theta_1'][::step].values.flatten()
        d['cost'] = estimator.blackbox_.epoch_log.get('train_cost')[::step]
        models[name] = d
        names.append(name)

    # ------------------------------------------------------------------  #
    # Obtain training data and set range along x axis
    X_train_ = estimators[names[0]].X_train_
    y_train_ = estimators[names[0]].y_train_
    xm, xM = np.min(X_train_[:,1]), np.max(X_train_[:,1])        
    xx = np.linspace(xm, xM)            

    # ------------------------------------------------------------------  #
    # Create data for surface plot
    theta0_min, theta0_max = min(theta0), max(theta0)
    theta1_min, theta1_max = min(theta1), max(theta1)
    theta0_mesh = np.linspace(theta0_min, theta0_max, 50)
    theta1_mesh = np.linspace(theta1_min, theta1_max, 50)
    theta0_mesh_grid, theta1_mesh_grid = np.meshgrid(theta0_mesh, theta1_mesh)    

    # ------------------------------------------------------------------  #
    # Create function to create mesh grid of costs 
    def cost_mesh(X, y, THETA):
        return(np.sum((X.dot(THETA) - y)**2)/(2*len(y)))          
    # Create z axis grid based upon X,y and the grid of thetas
    Js = np.array([cost_mesh(X_train_, y_train_, THETA)
                for THETA in zip(np.ravel(theta0_mesh_grid), np.ravel(theta1_mesh_grid))])
    Js = Js.reshape(theta0_mesh_grid.shape)          
    
    # ------------------------------------------------------------------  #
    # Create regression line data
    n_frames = len(models[names[0]]['theta_0'])

    def f(x, theta0, theta1):
        return theta0 + x * theta1
            
    for name, model in models.items():
        yy = []
        for j in range(n_frames):
            ym = f(xm, model['theta_0'][j], model['theta_1'][j])
            yM = f(xM, model['theta_0'][j], model['theta_1'][j])
            yy.append(np.linspace(ym, yM))
        models[name]['yy'] = yy

    # ------------------------------------------------------------------  #
    # Create subplots
    fig = make_subplots(rows=1, cols=2, subplot_titles=("Gradient Descent", "Linear Regression"),
                        specs=[[{'type': "surface"}, {"type": "xy"}]])      

    # ------------------------------------------------------------------  #
    # Add colors to model
    colors = ["red", "green", "black"]
    for i, model in enumerate(models.values()):
        model['color'] = colors[i]

    # ------------------------------------------------------------------  #
    # Add Surface Plot        
    # Subplot 1, Trace 0: Surface plot
    fig.add_trace(
        go.Surface(x=theta0_mesh, y=theta1_mesh, z=Js, colorscale="YlGnBu", 
                    showscale=False, showlegend=False), row=1, col=1)

    # ------------------------------------------------------------------  #
    # Subplots 1, Traces 1, 2 & 3: Gradient Descent Trajectories
    for name, model in models.items():
        fig.add_trace(
            go.Scatter3d(x=model['theta_0'][:1], y=model['theta_1'][:1], z=model['cost'][:1],
                        name=name, 
                        showlegend=False, 
                        mode='lines', line=dict(color=model['color'], width=5)),
                        row=1, col=1)            

    # ------------------------------------------------------------------  #
    # Subplot 2, Trace 4 Ames Data
    fig.add_trace(
        go.Scatter(x=X_train_[:,1], y=y_train_,
                    name="Ames Data",
                    mode="markers",
                    showlegend=True,
                    marker=dict(color="#1560bd")), row=1, col=2)                    

    # ------------------------------------------------------------------  #
    # Subplot 2, Traces 5, 6 & 7: Regression Lines
    for name, model in models.items():
        fig.add_trace(
            go.Scatter(x=xx, y=model['yy'][0], 
                    name=name,
                    mode="lines", marker=dict(color=model['color'], size=0.5)),
                    row=1, col=2)

    # ------------------------------------------------------------------  #
    # Set layout title, font, template, etc...
    fig.update_layout(
        height=600,
        #scene_xaxis=dict(range=[theta0_min, theta0_max], autorange=False),
        #scene_yaxis=dict(range=[theta1_min, theta1_max], autorange=False),            
        #scene_zaxis=dict(range=[zm, zM], autorange=False),
        title=dict(xanchor='center', yanchor='top', x=0.5, y=0.9),        
        font=dict(family="Open Sans"),                
        showlegend=True,            
        template='plotly_white');                       

    # ------------------------------------------------------------------  #
    # Create frames                       
    frames = [go.Frame(
        dict(
            name = f'{k+1}',
            data = [                    
                go.Scatter3d(x=models[names[0]]['theta_0'][:k+1], 
                                y=models[names[0]]['theta_1'][:k+1], 
                                z=models[names[0]]['cost'][:k+1]),
                go.Scatter3d(x=models[names[1]]['theta_0'][:k+1], 
                                y=models[names[1]]['theta_1'][:k+1], 
                                z=models[names[1]]['cost'][:k+1]),                                 
                go.Scatter3d(x=models[names[2]]['theta_0'][:k+1], 
                                y=models[names[2]]['theta_1'][:k+1], 
                                z=models[names[2]]['cost'][:k+1]),                                                                  
                go.Scatter(x=xx, y=models[names[0]]['yy'][k]),
                go.Scatter(x=xx, y=models[names[1]]['yy'][k]),
                go.Scatter(x=xx, y=models[names[2]]['yy'][k])
            ],
            traces=[1,2,3,5,6,7])
        ) for k in range(n_frames-1)]

    # Update the menus
    updatemenus = [dict(type='buttons',
                        buttons=[dict(label="Play",
                                        method="animate",
                                        args=[[f'{k+1}' for k in range(n_frames-1)],
                                        dict(frame=dict(duration=1, redraw=True),
                                                transition=dict(duration=1),
                                                easing="linear",
                                                fromcurrent=True,
                                                mode="immediate")])],
                        direction="left",
                        pad=dict(r=10, t=85),
                        showactive=True, x=0.1, y=0, xanchor="right", yanchor="top")]

    sliders = [{"yanchor": "top",
                "xanchor": "left",
                "currentvalue": {"font": {"size": 16}, "prefix": "Iteration: ", "visible":True, "xanchor": "right"},
                'transition': {'duration': 1, 'easing': 'linear'},
                'pad': {'b': 10, 't': 50}, 
                'len': 0.9, 'x': 0.1, 'y': 0, 
                'steps': [{'args': [[f'{k+1}'], {'frame': {'duration': 1, 'easing': 'linear', 'redraw': False},
                                    'transition': {'duration': 1, 'easing': 'linear'}}], 
                    'label': k, 'method': 'animate'} for k in range(n_frames-1)       
                ]}]

    fig.update(frames=frames)

    fig.update_layout(
        updatemenus=updatemenus,
        sliders=sliders
    )

    if filepath:
        fig.write_html(filepath, include_plotlyjs='cdn', include_mathjax='cdn')
    if show:
        pio.renderers.default = "browser"
        fig.show()

# --------------------------------------------------------------------------- #
#                             SingleModelSearch3D                             #  
# --------------------------------------------------------------------------- #
class SingleModelSearch3D:
    """Animates search for 2D optimization in 3D"""    

    def __init__(self):
        pass        

    def _cost_mesh(self,X, y, THETA):
        return(np.sum((X.dot(THETA) - y)**2)/(2*len(y)))        

    def search(self, model, directory=None, filename=None, fontsize=None,
                    interval=200, secs=10, maxframes=100):
        '''Animates gradient descent for 2D problems in 3D.
        '''        

        # Unpack model parameters
        params = model.get_params()   

        # Designate plot area
        fig = plt.figure(figsize=(12, 8))
        ax = fig.add_subplot(111, projection='3d')
        sns.set(style="whitegrid", font_scale=1)

        # Create index for n <= maxframes number of points
        iterations = np.arange(1, model.blackbox_.total_epochs+1)
        idx = np.arange(0,model.blackbox_.total_epochs)
        nth = math.floor(model.blackbox_.total_epochs/maxframes)
        nth = max(nth,1) 
        idx = idx[::nth]
        points = len(idx)

        # Create the x=theta0, y=theta1 grid for plotting
        weights = todf(model.blackbox_.epoch_log['theta'], stub='theta_')        
        theta0 = weights['theta_0']
        theta1 = weights['theta_1']

        # Format frames per second
        fps = math.floor(points/secs) if points >= secs else math.floor(secs/points)      

        # Establish boundaries of plot
        theta0_min = min(theta0)
        theta1_min = min(theta1)
        theta0_max = max(theta0) + abs(max(theta0)-min(theta0))
        theta1_max = max(theta1) + abs(max(theta1)-min(theta1))
        theta0_mesh = np.linspace(theta0_min, theta0_max, 50)
        theta1_mesh = np.linspace(theta1_min, theta1_max, 50)
        theta0_mesh, theta1_mesh = np.meshgrid(theta0_mesh, theta1_mesh)

        # Create cost grid based upon X,y and the grid of thetas
        Js = np.array([self._cost_mesh(model.X, model.y, THETA)
                    for THETA in zip(np.ravel(theta0_mesh), np.ravel(theta1_mesh))])
        Js = Js.reshape(theta0_mesh.shape)

        # Set Title
        plt.rc('text', usetex=True)
        title = model.name + '\n' + r'$\alpha$' + " = " + str(round(params['learning_rate'],3))
        ax.set_title(title, color='k', pad=30, fontsize=20)                            
        if fontsize:            
            display = ax.text2D(0.1,0.92, '', transform=ax.transAxes, color='k', fontsize=fontsize)
        else:            
            display = ax.text2D(0.2,0.92, '', transform=ax.transAxes, color='k')             
        # Set face, tick,and label colors 
        ax.set_facecolor('w')
        ax.tick_params(colors='k')
        ax.xaxis.label.set_color('k')
        ax.yaxis.label.set_color('k')
        # make the panes transparent
        ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        # make the grid lines transparent
        ax.xaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        ax.yaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        ax.zaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        # Make surface plot
        ax.plot_surface(theta0_mesh, theta1_mesh, Js, rstride=1,
                cstride=1, cmap='jet', alpha=0.5, linewidth=0)
        ax.set_xlabel(r'Intercept($\theta_0$)')
        ax.set_ylabel(r'Slope($\theta_1$)')
        ax.set_zlabel('Cost')        
        ax.view_init(elev=30., azim=30)

        # Build the empty line plot at the initiation of the animation
        line2d, = ax.plot([], [], [], 'b-', lw = 1.5)
        line3d, = ax.plot([], [], [], 'r-', lw = 1.5)        
        point2d, = ax.plot([], [], [], 'bo')
        point3d, = ax.plot([], [], [], 'ro')

        def init():

            # Initialize 3d line and point
            line3d.set_data([],[])
            line3d.set_3d_properties([])
            point3d.set_data([], [])
            point3d.set_3d_properties([])

            # Initialize 2d line and point
            line2d.set_data([],[])
            line2d.set_3d_properties([])
            point2d.set_data([], [])
            point2d.set_3d_properties([])            

            # Initialize display
            display.set_text('')
            return (line2d, point2d, line3d, point3d, display,)

        # Animate the optimization path as it converges
        def animate(i):
            # Animate 3d Line
            line3d.set_data(theta0[:idx[i]], theta1[:idx[i]])
            line3d.set_3d_properties(model.blackbox_.epoch_log['train_cost'][:idx[i]])
            # Animate 3d points
            point3d.set_data(theta0[idx[i]], theta1[idx[i]])
            point3d.set_3d_properties(model.blackbox_.epoch_log['train_cost'][idx[i]])
            # Animate 2d Line
            line2d.set_data(theta0[:idx[i]], theta1[:idx[i]])
            line2d.set_3d_properties(0)
            # Animate 2d points
            point2d.set_data(theta0[idx[i]], theta1[idx[i]])
            point2d.set_3d_properties(0)            
            # Rotate
            ax.view_init(elev=30., azim=i*.30)
            # Update text
            metrics = 'Iteration: '+ str(iterations[idx[i]]) + \
                      r'$\quad\theta_0=$ ' + str(round(theta0[idx[i]],3)) \
                          + r'$\quad\theta_1=$ ' + str(round(theta1[idx[i]],3)) +\
                            '  Cost: ' + \
                                str(np.round(model.blackbox_.epoch_log['train_cost'][idx[i]], 3))
            display.set_text(metrics)            

            return(line3d, point3d, line2d, point2d, display)

        # create animation using the animate() function
        surface_ani = animation.FuncAnimation(fig, animate, init_func=init, frames=len(idx),
                                            interval=interval, blit=True, repeat_delay=3000)
        if directory is not None:
            if filename is None:
                filename = model.algorithm + ' Search Path Learning Rate ' + str(params['learning_rate']) +  '.gif'
            save_gif(surface_ani, directory, filename, fps)
        return(surface_ani)

# --------------------------------------------------------------------------- #
#                              SingleModelFit2D                               #   
# --------------------------------------------------------------------------- #
class SingleModelFit2D:
    """Animates gradient descent fit in 2D for 2D optimation problem."""

    def fit(self, model, directory=None, filename=None, fontsize=None,
                 interval=200, secs=10, maxframes=100):

        # Unpack model results
        params = model.get_params()         
        X = model.X_train_
        y = model.y_train_        

        # Create index for n <= maxframes number of points
        iterations = np.arange(1, model.blackbox_.total_epochs+1)
        idx = np.arange(0,model.blackbox_.total_epochs)
        nth = math.floor(model.blackbox_.total_epochs/maxframes)
        nth = max(nth,1) 
        idx = idx[::nth]
        points = len(idx)

        # Extract data for plotting
        x = model.X[:,1]        
        cost = model.blackbox_.epoch_log['train_cost']        
        weights = todf(model.blackbox_.epoch_log['theta'], stub='theta_')        
        theta0 = weights['theta_0']
        theta1 = weights['theta_1']
        theta = np.array([theta0, theta1])

        # Format frames per second
        fps = math.floor(points/secs) if points >= secs else math.floor(secs/points)      
        
        # Render scatterplot
        fig, ax = plt.subplots(figsize=(12,8))
        sns.set(style="whitegrid", font_scale=1)
        sns.scatterplot(x=x, y=y, ax=ax)
        ax.get_yaxis().set_major_formatter(
            mpl.ticker.FuncFormatter(lambda x, p: format(int(x), ',')))
    
        # Set face, tick,and label colors 
        ax.set_facecolor('w')
        ax.tick_params(colors='k')
        ax.xaxis.label.set_color('k')
        ax.yaxis.label.set_color('k')
        # Initialize line
        line, = ax.plot([],[],'r-', lw=2)
        # Set Title, Annotations and label
        title = model.algorithm + '\n' + r'$\alpha$' + " = " + str(round(params['learning_rate'],3))
        ax.set_title(title, color='k', fontsize=20)
        if fontsize:            
            display = ax.text(0.1, 0.9, '', transform=ax.transAxes, color='k', fontsize=fontsize)
        else:            
            display = ax.text(0.2, 0.9, '', transform=ax.transAxes, color='k')
        ax.set_xlabel('X')
        ax.set_ylabel('y')
        fig.tight_layout()

        # Build the empty line plot at the initiation of the animation
        def init():
            line.set_data([],[])
            display.set_text('')
            return (line, display,)

        # Animate the regression line as it converges
        def animate(i):

            # Animate Line
            y=X.dot(theta[:,idx[i]])
            line.set_data(x,y)

            # Animate text
            display.set_text('Iteration: '+ str(iterations[idx[i]]) + r'$\quad\theta_0=$ ' +
                            str(round(theta0[idx[i]],3)) + r'$\quad\theta_1=$ ' + str(round(theta1[idx[i]],3)) +
                            ' Cost: ' + str(round(cost[idx[i]], 3)))
            return (line, display)

        # create animation using the animate() function
        line_gd = animation.FuncAnimation(fig, animate, init_func=init, frames=len(idx),
                                            interval=interval, blit=True, repeat_delay=3000)
        if directory is not None:
            if filename is None:
                filename = model.algorithm + ' Fit Plot Learning Rate ' + str(params['learning_rate']) +  '.gif'
            save_gif(line_gd, directory, filename, fps)
        plt.close(fig)  
        return(line_gd)

# --------------------------------------------------------------------------- #
#                            MultiModelSearch3D                          #  
# --------------------------------------------------------------------------- #
class MultiModelSearch3D(animation.FuncAnimation):
    """Animates gradient descent in 3D for multiple 2D optimization problems."""

    def __init__(self):
        pass

    def _anim8(self, *paths, zpaths, methods=[], frames=100, 
                 interval=60, repeat_delay=5, blit=False, **kwargs):

        self.paths = paths
        self.zpaths = zpaths

        if frames is None:
            frames = max(path.shape[1] for path in paths)
            self.nth = [1 for p in paths]
        else:
            self.nth = [max(1,math.floor(path.shape[1] / frames)) for path in paths]
        cmap = plt.get_cmap('jet')
        colors = cmap(np.linspace(0,1,len(paths)))
        self.lines = [self.ax.plot([], [], [], alpha=0.7, label=method, c=c, lw=2)[0] 
                      for _, method, c in zip_longest(paths, methods, colors)]

        ani = animation.FuncAnimation(self.fig, self._update, init_func=self._init,
                                      frames=frames, interval=interval, blit=blit,
                                      repeat_delay=repeat_delay, **kwargs)
        self.ax.legend(loc='upper left')                                      
        return ani

    def _init(self):
        for line in self.lines:
            line.set_data([], [])
            line.set_3d_properties([])
        return self.lines

    def _update(self, i):
        for line, nth, path, zpath in zip(self.lines, self.nth, self.paths, self.zpaths):
            self.ax.view_init(elev=60., azim=i*.30)
            line.set_data(*path[::,:i*nth])
            line.set_3d_properties(zpath[:i*nth])
        return self.lines

    def _cost_mesh(self,X, y, THETA):
        return(np.sum((X.dot(THETA) - y)**2)/(2*len(y))) 

    def _meshgrid(self, models):
        theta0_mins = []
        theta1_mins = []
        theta0_maxs = []
        theta1_maxs = []

        for _, v in models.items():
            theta0_mins.append(min(item[0] for item in v.blackbox_.batch_log.get('theta')))
            theta1_mins.append(min(item[1] for item in v.blackbox_.batch_log.get('theta')))
            theta0_maxs.append(max(item[0] for item in v.blackbox_.batch_log.get('theta')))
            theta1_maxs.append(max(item[1] for item in v.blackbox_.batch_log.get('theta')))
        
        theta0_min = min(theta0_mins)
        theta1_min = min(theta1_mins)
        theta0_max = max(theta0_maxs)
        theta1_max = max(theta1_maxs)

        theta0_mesh = np.linspace(theta0_min, theta0_max, 100)
        theta1_mesh = np.linspace(theta1_min, theta1_max, 100)
        theta0_mesh, theta1_mesh = np.meshgrid(theta0_mesh, theta1_mesh)
        return theta0_mesh, theta1_mesh

    def _surface(self, X, y, models, height, width):

        # Designate plot area
        self.fig = plt.figure(figsize=(12*width, 8*height))
        self.ax = self.fig.add_subplot(111, projection='3d')
        sns.set(style="whitegrid", font_scale=1)        

        # Create cost grid based upon x,y the grid of thetas
        theta0_mesh, theta1_mesh = self._meshgrid(models)
        Js = np.array([self._cost_mesh(X, y, THETA) 
                    for THETA in zip(np.ravel(theta0_mesh), np.ravel(theta1_mesh))])
        Js = Js.reshape(theta0_mesh.shape)

        # Set Title
        title = 'Gradient Descent Trajectories'
        self.ax.set_title(title, color='k', pad=30, fontsize=20)                       

        # Set face, tick,and label colors 
        self.ax.set_facecolor('w')
        self.ax.tick_params(colors='k')
        self.ax.xaxis.label.set_color('k')
        self.ax.yaxis.label.set_color('k')
        # make the panes transparent
        self.ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        self.ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        self.ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        # make the grid lines transparent
        self.ax.xaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        self.ax.yaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        self.ax.zaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        # Make surface plot
        self.ax.plot_surface(theta0_mesh, theta1_mesh, Js, rstride=1,
                cstride=1, cmap='jet', alpha=0.5, linewidth=0)
        self.ax.set_xlabel(r'Bias($\theta_0$)')
        self.ax.set_ylabel(r'Coefficient($\theta_1$)')
        self.ax.set_zlabel('Cost')        
        self.ax.view_init(elev=30., azim=30)        

    def _get_data(self, models):
        
        paths=[]
        zpaths=[]
        methods = []        
        for k, v in models.items():    
            paths.append(np.array(v.blackbox_.epoch_log.get('theta')).T)
            zpaths.append(np.array(v.blackbox_.epoch_log.get('train_cost'))) 
            methods.append(k)  
            X = v.X_train_
            y = v.y_train_
        return(X, y, paths, zpaths, methods)

    def search(self, models, frames=100, directory=None, filename=None, fps=5,
               height=1, width=1):

        X, y, paths, zpaths, methods = self._get_data(models)        
        self._surface(X, y, models, height, width)        
        ani = self._anim8(*paths, zpaths=zpaths, methods=methods, frames=frames)
        if directory is not None:
            if filename is None:
                filename = 'Gradient Descent Trajectories.gif'
            save_gif(ani, directory, filename, fps)
        return ani


# --------------------------------------------------------------------------- #
#                           GRADIENT FIT 2D CLASS                             #   
# --------------------------------------------------------------------------- #
class MultiModelFit2D(animation.FuncAnimation):
    """Animates fit for multiple 2D optimizations.""" 

    def __init__(self):
        pass

    def _anim8(self, *paths, methods=[], frames=None, 
                 interval=60, repeat_delay=5, blit=False, **kwargs):

        self.paths = paths

        if frames is None:
            frames = max(path.shape[1] for path in paths)
            self.nth = [1 for p in paths]
        else:
            self.nth = [max(1,math.floor(path.shape[1] / frames)) for path in paths]
        cmap = plt.get_cmap('jet')
        colors = cmap(np.linspace(0,1,len(paths)))
        self.lines = [self.ax.plot([], [], label=method, c=c, lw=2)[0] 
                      for _, method, c in zip_longest(paths, methods, colors)]

        ani = animation.FuncAnimation(self.fig, self._update, init_func=self._init,
                                      frames=frames, interval=interval, blit=blit,
                                      repeat_delay=repeat_delay, **kwargs)
        self.ax.legend(loc='upper left')                                      
        return(ani)                                                  

    def _init(self):
        for line in self.lines:
            line.set_data([], [])
        return self.lines

    def _update(self, i):
        for line, nth, path, in zip(self.lines, self.nth, self.paths):
            y = self.X.dot(path[::,i*nth])
            line.set_data(self.X[:,1], y)            
        return self.lines

    def _scatterplot(self, x_label=None, y_label=None):

        # Designate plot area
        self.fig, self.ax = plt.subplots(figsize=(12,8))
        sns.set(style="whitegrid", font_scale=1)
        sns.scatterplot(x=self.X[:,1], y=self.y, ax=self.ax)

        # Set face, tick,and label colors 
        self.ax.set_facecolor('w')
        self.ax.tick_params(colors='k')
        self.ax.xaxis.label.set_color('k')
        self.ax.yaxis.label.set_color('k')

        # Set Title, Annotations and label
        title = 'Gradient Descent Model Fit'
        self.ax.set_title(title, color='k', fontsize=20)
        self.ax.set_xlabel(x_label) if x_label is not None else self.ax.set_xlabel("X")
        self.ax.set_ylabel(y_label) if x_label is not None else self.ax.set_ylabel("y")
        self.fig.tight_layout()      

    def _get_data(self, models):
        
        paths=[]
        methods = []
        for k, v in models.items():    
            paths.append(np.array(v.blackbox_.batch_log.get('theta')).T)
            methods.append(k)  
            self.X = v.X_train_
            self.y = v.y_train_

        return paths, methods
    def fit(self, models, frames=100, x_label=None, y_label=None, 
            directory=None, filename=None, fps=5):

        paths, methods = self._get_data(models)
        self._scatterplot(x_label, y_label)
        ani = self._anim8(*paths, methods=methods, frames=frames)
        if directory is not None:
            if filename is None:
                filename = 'Gradient Descent Fit.gif'            
            save_gif(ani, directory, filename, fps)
        return(ani)        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : activation.py                                                     #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Saturday, May 16th 2020, 11:39:45 pm                        #
# Last Modified : Saturday, May 16th 2020, 11:39:45 pm                        #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Activation functions."""
from abc import ABC, abstractmethod

import numpy as np
from sklearn.base import BaseEstimator

from mlstudio.utils.validation import validate_range
# --------------------------------------------------------------------------  #
# Functions adapted from the following sources.
# Reference https://en.wikipedia.org/wiki/Activation_function 
# Reference https://www.python-course.eu/softmax.php
# --------------------------------------------------------------------------  #
class Activation(ABC, BaseEstimator):
    """Base class for activation classes."""
    @abstractmethod
    def __call__(self, x):
        pass

    @abstractmethod
    def gradient(self, x):
        pass

class Sigmoid(Activation):
    """Sigmoid activation function."""

    def __call__(self, x):
        if np.all(x >= 0):            
            s = 1 / (1 + np.exp(-x))
        else:
            s =  np.exp(x) / (1 + np.exp(x))    
        return s
        

    def gradient(self, x):
        return self.__call__(x) * (1 - self.__call__(x))

class Softmax(Activation):
    def __call__(self, x):
        # Subtract the max to avoid underflow or overflow errors 
        e_x = np.exp(x - np.max(x, axis=-1, keepdims=True))
        s = e_x / np.sum(e_x, axis=-1, keepdims=True)
        return s

    def gradient(self, x):
        p = self.__call__(x)
        return p * (1 - p)

class TanH(Activation):
    def __call__(self, x):
        return (np.exp(x) - np.exp(-x)) / (np.exp(x) + np.exp(-1))        

    def gradient(self, x):
        return 1 - np.power(self.__call__(x), 2)

class ReLU(Activation):
    def __call__(self, x):
        return np.where(x > 0, x, 0)

    def gradient(self, x):
        return np.where(x > 0, 1, 0)

class LeakyReLU(Activation):
    def __init__(self, alpha=0.01):
        self.alpha = alpha

    def __call__(self, x):
        return np.where(x >= 0, x, self.alpha * x)

    def gradient(self, x):
        return np.where(x >= 0, 1, self.alpha)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : MLStudio                                                          #
# Version : 0.1.0                                                             #
# File    : variants.py                                                       #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, April 10th 2020, 1:52:41 pm                         #
# Last Modified : Friday, April 10th 2020, 1:52:41 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
#%%
from pathlib import Path
import site
PROJECT_DIR = Path(__file__).resolve().parents[3]
site.addsitedir(PROJECT_DIR)

import matplotlib.pyplot as plt
import numpy as np
from sklearn.datasets import make_regression

from mlstudio.supervised.estimator.gradient import GradientDescentRegressor
from mlstudio.utils.data_manager import StandardScaler
from mlstudio.visual.animations.gradient import MultiModelSearch3D
from mlstudio.visual.animations.gradient import MultiModelFit2D

# Obtain and standardize data
X, y, coef = make_regression(n_samples=1000, n_features=1, effective_rank=5, 
                             noise=50, random_state=5, coef=True)
scaler = StandardScaler()
X = scaler.transform(X)

# Run Models
name = np.array(['Batch Gradient Descent', 'Minibatch Gradient Descent', 'Stochastic Gradient Descent'])
theta_init = np.array([-15., 15.])
batch_size = np.array([None, 32, 1])
models = {}
for i in range(len(batch_size)):
    models[name[i]] = GradientDescentRegressor(theta_init=theta_init, batch_size=batch_size[i]).fit(X,y)

# Run visualizations and save
directory = "../figures/variants/"
filename = "gradient_descent_search.gif"
viz = MultiModelSearch3D()
viz.search(models=models, directory=directory, filename=filename)
plt.show()
filename = "gradient_descent_fit.gif"
viz = MultiModelFit2D()
viz.fit(models=models, directory=directory, filename=filename)
plt.show()




# %%



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.14                                                            #
# File    : early_stop.py                                                     #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Friday, May 15th 2020, 9:16:41 pm                           #
# Last Modified : Sunday, June 14th 2020, 11:49:38 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Early stop classes."""
from abc import ABC, abstractmethod
import datetime
import numpy as np
import pandas as pd

from mlstudio.supervised.observers.base import Observer, PerformanceObserver
from mlstudio.utils.validation import validate_metric, validate_int
from mlstudio.utils.validation import validate_zero_to_one
# --------------------------------------------------------------------------- #
#                               PERFORMANCE                                   #
# --------------------------------------------------------------------------- #
class Performance(PerformanceObserver):
    """Performances and log model performance, critical points and stability. 

    Performance is defined in terms of:

        * Metric : A metric to observe. This can be training error, validation
            score, gradient norm or the like 

        * Epsilon : A mininum amount of relative change in the observed
            metric required to consider the optimization in a productive
            state.

        * Patience : The number of consecutive epochs or iterations of
            non-improvement that is tolerated before considering an
            optimization stabilized.    

    Parameters
    ----------
    mode : str 'active' or 'passive' (Default='passive')
        In 'active' mode, this observer signals the estimator to suspend
        optimization when performance hasn't improved. In 'passive' mode,
        the observer collects, analyzes and stores performance data, but
        does not effect the subject's behavior. 

    val_size : float (default=0.3)
        The proportion of the training set to allocate to the 
        validation set.

    metric : str, optional (default='train_cost')
        Specifies which statistic to metric for evaluation purposes.

        'train_cost': Training set costs
        'train_score': Training set scores based upon the model's metric parameter
        'val_cost': Validation set costs
        'val_score': Validation set scores based upon the model's metric parameter
        'gradient_norm': The norm of the gradient of the objective function w.r.t. theta

    epsilon : float, optional (default=0.01)
        The amount of relative change in the observed metric considered to be
        a sufficient improvement in performance. 

    patience : int, optional (default=5)
        The number of consecutive epochs of non-improvement that is 
        tolerated before considering the optimization stable.

    best_or_final : str, optional (default='best')
        Indicates whether the composing estimator should use the best
        or final results from the Performance object when reporting results
        and computing predictions

    All estimator performance considerations are managed and controlled
    by this class. 
    """

    def __init__(self, mode='passive', val_size=0.3, metric='val_score', 
                 epsilon=0.001, patience=5, best_or_final='best'): 
        super(Performance, self).__init__(                   
            metric = metric,
            epsilon = epsilon,
            patience = patience
        )
        self.name = "Performance EarlyStop Observer"
        self.mode = mode       
        self.best_or_final = best_or_final 
        self.val_size = val_size

    def on_train_begin(self, log=None):
        """Logic executed at the beginning of training.
        
        Parameters
        ----------
        log: dict
            Currently not used
        """
        super(Performance, self).on_train_begin(log=log)
        self.model.val_size = self.val_size                          

    def on_epoch_end(self, epoch, log=None):
        """Logic executed at the end of each epoch.
        
        Parameters
        ----------
        epoch : int
            Current epoch
        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """                  
        super(Performance, self).on_epoch_end(epoch=epoch, log=log)
        if self._stabilized and self.mode == 'active':
            self.model.converged = True       

    def on_train_end(self, log=None):
        """Logic executed at the end of training.

        Updates the model with the best results and the critical points.
        
        Parameters
        ----------        
        log: dict
            Dictionary containing the data, cost, batch size and current weights
        """           
        self.model.best_results_ = self._best_results        
        self.model.critical_points_ = self._critical_points




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : static.py                                                         #
# Python  : 3.8.3                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Wednesday, June 10th 2020, 8:34:33 pm                       #
# Last Modified : Wednesday, June 10th 2020, 8:34:33 pm                       #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Module for static plots"""
from collections import OrderedDict
import os
from pathlib import Path
import sys

import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go 
import plotly.io as pio 
import plotly.offline as py
from plotly.subplots import make_subplots
import seaborn as sns

from mlstudio.utils.file_manager import check_directory

def plot_cost_scores(data, filepath=None, show=True):
    """Renders subplots of the optimization and regression line fit.
    
    Parameters
    ----------
    data : dict
        Dictionary containing an error and a score dictionary. Each of
        those dictionaries contain the errors and scores for 
        one or more estimators. 

    filepath : str (default=None)
        The path to which the animation is to be saved.

    show : bool (default=True)
        Indicates whether an animation should be rendered in users browser.

    """        
    # ------------------------------------------------------------------  #
    # Extract plot data
    cost_plot_data = data['cost']
    score_plot_data = data['score']

    # ------------------------------------------------------------------  #
    # Get colors
    colors = px.colors.qualitative.Prism        

    # ------------------------------------------------------------------  #
    # Create subplots
    title_error = "Training Error (MSE)"
    title_score = "Test Scores (" + r'$R^2$' + ")"
    fig = make_subplots(rows=1, cols=2, subplot_titles=(title_error, title_score),
                        specs=[[{'type': "xy"}, {"type": "xy"}]])      

    # ------------------------------------------------------------------  #
    # Add Training Error Plot
    # Subplot 1, Trace 0: Bar Plot
    fig.add_trace(
        go.Bar(x=list(cost_plot_data.keys()), y=list(cost_plot_data.values()), 
                marker=dict(color=colors), showlegend=False),
                row=1, col=1)

    # ------------------------------------------------------------------  #
    # Add Test Score Plot
    # Subplot 2, Trace 1: Bar Plot
    fig.add_trace(
        go.Bar(x=list(score_plot_data.keys()), y=list(score_plot_data.values()), 
                marker=dict(color=colors), showlegend=False,
                text=[round(num,3) for num in list(score_plot_data.values())], 
                textposition='auto'),
                row=1, col=2)
    # ------------------------------------------------------------------  #
    # Set layout title, font, template, etc...
    fig.update_layout(
        height=600,
        title=dict(xanchor='center', yanchor='top', x=0.5, y=0.9),        
        font=dict(family="Open Sans"),                
        showlegend=True,            
        template='plotly_white');                       

    # ------------------------------------------------------------------  #
    # Save and rendor plot
    if filepath:
        fig.write_html(filepath, include_plotlyjs='cdn', include_mathjax='cdn')
    if show:
        pio.renderers.default = "browser"
        fig.show()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# =========================================================================== #
# Project : ML Studio                                                         #
# Version : 0.1.0                                                             #
# File    : linear_regression.py                                              #
# Python  : 3.8.2                                                             #
# --------------------------------------------------------------------------  #
# Author  : John James                                                        #
# Company : DecisionScients                                                   #
# Email   : jjames@decisionscients.com                                        #
# URL     : https://github.com/decisionscients/MLStudio                       #
# --------------------------------------------------------------------------  #
# Created       : Tuesday, May 19th 2020, 5:19:25 pm                          #
# Last Modified : Tuesday, May 19th 2020, 5:19:25 pm                          #
# Modified By   : John James (jjames@decisionscients.com)                     #
# --------------------------------------------------------------------------  #
# License : BSD                                                               #
# Copyright (c) 2020 DecisionScients                                          #
# =========================================================================== #
"""Regression algorithms.
This class encapsulates the core behaviors for regression classes. Currently,
the following regression classes are supported.
    
    * Linear Regression
    * Lasso Regression
    * Ridge Regression
    * ElasticNet Regression
The core behaviors exposed for each class include:
    * predict : Predicts outputs as linear combination of inputs and weights.
    * compute_cost : Computes cost associated with predictions
    * compute_gradient : Computes the derivative of loss w.r.t. to weights
"""
from abc import ABC, abstractmethod
import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_array

from mlstudio.supervised.machine_learning.base import BaseRegressor
# --------------------------------------------------------------------------- #
#                          LINEAR REGRESSION                                  #
# --------------------------------------------------------------------------- #    
class LinearRegression(BaseRegressor, RegressorMixin):
    """Linear Regression algorithm."""

    def __init__(self):
        pass

    @property
    def name(self):
        return "Linear Regression"    

    @property
    def task(self):
        return "Regression"

    def compute_output(self, X, theta):
        """Computes output based upon inputs and model parameters.
        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The model inputs. Note the number of features includes the coefficient
            for the bias term
        theta : array of shape [n_features,] or [n_features, n_classes]
            Model parameters
        Returns
        -------
        output : Model output            
        
        """
        return X.dot(theta)

    def predict(self, X, theta):
        """Computes the prediction as linear combination of inputs and parameters.        
        Parameter
        ---------
        X : array of shape [n_samples, n_features]
            The model inputs. 
        theta : array of shape [n_features,] 
            Model parameters
        Note: n_features may or may not include the bias term added prior to 
        training, so we will need to accommodate X of either dimension.
        Returns
        -------
        prediction : Linear combination of inputs.
        """         
        if X.shape[1] == len(theta):
            y_pred = X.dot(theta)
        else:
            y_pred = theta[0] + X.dot(theta[1:])
        return y_pred

    def compute_cost(self, y, y_out, theta):
        """Computes the mean squared error cost.
        Parameters
        ----------
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters            
        Returns
        -------
        cost : The quadratic cost 
        """
        J = np.mean(0.5 * (y-y_out)**2)
        return J

    def compute_gradient(self, X, y, y_out, theta):
        """Computes quadratic costs gradient with respect to weights.
        
        Parameters
        ----------
        X : array of shape (m_observations, n_features)
            Input data
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters                        
        Returns
        -------
        gradient of the cost function w.r.t. the parameters.
        """
        n_samples = X.shape[0]
        dZ = y_out-y
        dW = float(1./n_samples) * X.T.dot(dZ) 
        return(dW)           

# --------------------------------------------------------------------------- #
#                          LASSO REGRESSION                                   #
# --------------------------------------------------------------------------- #    
class LassoRegression(LinearRegression):
    """Lasso Regression algorithm."""
    
    def __init__(self, lambda_reg=0.0001):
        self.lambda_reg = lambda_reg

    @property
    def name(self):
        return "Lasso Regression"        

    def compute_cost(self, y, y_out, theta):
        """Computes the mean squared error cost.
        Parameters
        ----------
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters            
        Returns
        -------
        cost : The quadratic cost         
        """
        self._validate_hyperparam(self.lambda_reg)
        n_samples = y.shape[0]
        J_reg = (self.lambda_reg / n_samples) * np.linalg.norm(theta, ord=1)
        J = np.mean(0.5 * (y-y_out)**2) + J_reg
        return J

    def compute_gradient(self, X, y, y_out, theta):
        """Computes quadratic costs gradient with respect to weights.
        
        Parameters
        ----------
        X : array of shape (m_observations, n_features)
            Input data
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters                        
        Returns
        -------
        gradient of the cost function w.r.t. the parameters.
        """
        n_samples = X.shape[0]
        dZ = y_out-y
        dW = 1/n_samples  * (X.T.dot(dZ) + self.lambda_reg * np.sign(theta))
        return(dW)           

# --------------------------------------------------------------------------- #
#                          RIDGE REGRESSION                                   #
# --------------------------------------------------------------------------- #            
class RidgeRegression(LinearRegression):
    """Ridge Regression algorithm."""
    
    def __init__(self, lambda_reg=0.0001):
        self.lambda_reg=lambda_reg

    @property
    def name(self):
        return "Ridge Regression"

    def compute_cost(self, y, y_out, theta):
        """Computes the mean squared error cost.
        Parameters
        ----------
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters            
        Returns
        -------
        cost : The quadratic cost 
        """
        self._validate_hyperparam(self.lambda_reg)
        n_samples = y.shape[0]
        J_reg = (self.lambda_reg / (2*n_samples)) * np.linalg.norm(theta)**2
        J = np.mean(0.5 * (y-y_out)**2) + J_reg
        return J

    def compute_gradient(self, X, y, y_out, theta):
        """Computes quadratic costs gradient with respect to weights.
        
        Parameters
        ----------
        X : array of shape (m_observations, n_features)
            Input data
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters                        
        Returns
        -------
        gradient of the cost function w.r.t. the parameters.
        """
        n_samples = X.shape[0]
        dZ = y_out-y
        dW = 1/n_samples  * (X.T.dot(dZ) + self.lambda_reg * theta)
        return(dW)                         

# --------------------------------------------------------------------------- #
#                        ELASTIC NET REGRESSION                               #
# --------------------------------------------------------------------------- #            
class ElasticNetRegression(LinearRegression):
    """Elastic Net Regression algorithm."""
    
    def __init__(self, lambda_reg=0.0001, ratio=0.15):
        self.lambda_reg=lambda_reg
        self.ratio=ratio

    @property
    def name(self):
        return "ElasticNet Regression"               

    def compute_cost(self, y, y_out, theta):
        """Computes the mean squared error cost.
        Parameters
        ----------
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters            
        Returns
        -------
        cost : The quadratic cost 
        """
        n_samples = y.shape[0]
        self._validate_hyperparam(self.lambda_reg)
        self._validate_hyperparam(self.ratio)
        l1_contr = self.ratio * np.linalg.norm(theta, ord=1)
        l2_contr = (1 - self.ratio) * 0.5 * np.linalg.norm(theta)**2        
        J_reg = float(1./n_samples) * self.lambda_reg * (l1_contr + l2_contr)
        J = np.mean(0.5 * (y-y_out)**2) + J_reg
        return J

    def compute_gradient(self, X, y, y_out, theta):
        """Computes quadratic costs gradient with respect to weights.
        
        Parameters
        ----------
        X : array of shape (m_observations, n_features)
            Input data
        y : array of shape (n_features,)
            Ground truth target values
        y_out : array of shape (n_features,)
            Output from the model 
        theta : array of shape (n_features,)  
            The model parameters                        
        Returns
        -------
        gradient of the cost function w.r.t. the parameters.
        """
        n_samples = X.shape[0]
        l1_contr = self.ratio * np.sign(theta)
        l2_contr = (1 - self.ratio) * theta
        lambda_reg = np.asarray(self.lambda_reg, dtype='float64')         
        dZ = y_out-y
        dW = 1/n_samples  * (X.T.dot(dZ) + np.multiply(lambda_reg, np.add(l1_contr, l2_contr)))
        return(dW)     


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# ============================================================================ #
# Project : MLStudio                                                           #
# Version : 0.1.0                                                              #
# File    : base.py                                                            #
# Python  : 3.8.2                                                              #
# ---------------------------------------------------------------------------- #
# Author  : John James                                                         #
# Company : DecisionScients                                                    #
# Email   : jjames@decisionscients.com                                         #
# URL     : https://github.com/decisionscients/MLStudio                        #
# ---------------------------------------------------------------------------- #
# Created       : Tuesday, March 17th 2020, 7:15:23 pm                         #
# Last Modified : Tuesday, March 17th 2020, 7:15:23 pm                         #
# Modified By   : John James (jjames@decisionscients.com)                      #
# ---------------------------------------------------------------------------- #
# License : BSD                                                                #
# Copyright (c) 2020 DecisionScients                                           #
# ============================================================================ #
"""Base class for visualizations."""
import os
from abc import ABC, abstractmethod

import plotly.express as px
import plotly.graph_objects as go
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_X_y, check_array
# --------------------------------------------------------------------------- #
#                            VISUALATRIX                                      #
# --------------------------------------------------------------------------- #

class Visualatrix(ABC, BaseEstimator):
    """Abstact base class at the top of the visualator object hierarchy
    
    This base class defines how the interface and common behaviors for and
    among the visualizations classes.  

    Parameters
    ----------
    kwargs : dict
        Keyword arguments including:

        =========   ==========================================
        Property    Description
        --------    ------------------------------------------
        fig         Plotly Figure or FigureWidget object
        height      specify the height of the figure
        width       specify the width of the figure
        template    specify the template for the figure.
        title       specify the title for the figure.
        =========   ==========================================

    """
    _PLOT_DEFAULT_HEIGHT = 450
    _PLOT_DEFAULT_WIDTH  = 900   
    _PLOT_DEFAULT_TEMPLATE = "plotly_white"    
    _PLOT_AVAILABLE_TEMPLATES = ['ggplot2', 'seaborn', 'simple_white', 'plotly',
         'plotly_white', 'plotly_dark', 'presentation', 'xgridoff',
         'ygridoff', 'gridon', 'none']    
    
    def __init__(self, fig=None, **kwargs):        
        self.fig = fig
        self.height = kwargs.get('height', self._PLOT_DEFAULT_HEIGHT)
        self.width = kwargs.get('width', self._PLOT_DEFAULT_WIDTH)
        self.template = kwargs.get('template', self._PLOT_DEFAULT_TEMPLATE)
        self.title = kwargs.get('title', None)

    @abstractmethod
    def fit(self, X, y=None, **kwargs):
        """Fits the visualator to the data."""
        pass    

    def _validate(self, X, y=None, **kwargs):
        """Validates arguments passed to the fit method.
        
        Parameters
        ----------
        X : array-like of shape (m samples, n features)
            Object containing typically feature data.

        y : array-like of shape (m samples) (Optional)
            The target variable.
        """
        if y is not None:
            check_X_y(X, y)
        else:
            check_array(X)
        

    
    def show(self, **kwargs):
        """Renders the visualization"""
        self.fig.show()

    def save(self, filepath):
        """Saves image to filepath

        Parameters
        ----------
        filepath : str
            Relative filepath including file name and extension
        """
        directory = os.path.basename(filepath)
        if not os.path.exists(directory):
            os.mkdir(directory)
        self.fig.write_image(filepath)


# --------------------------------------------------------------------------- #
#                            ModelVisualatrix                                 #
# --------------------------------------------------------------------------- #

class ModelVisualatrix(Visualatrix):
    """Base class for model visualization classes."""

    def __init__(self, estimator, fig=None, **kwargs):
        super(ModelVisualatrix, self).__init__(fig=fig, **kwargs)
        self.estimator = estimator

    def _validate(self, X, y=None, **kwargs):
        """Validates parameters and arguments sent to the fit method.
        
        Parameters
        ----------
        X : array-like of shape (m samples, n features)
            Object containing typically feature data.

        y : array-like of shape (m samples) (Optional)
            The target variable.

        Raises
        ------
        ValueError : estimator has no 'fit' or 'predict' method.

        """
        super(ModelVisualatrix, self)._validate(X, y)

        # Confirm the estimator has fit and predict methods.
        if getattr(self.estimator, 'fit', None) and \
        getattr(self.estimator, 'predict', None):
            pass
        else:
            raise ValueError("The estimator object must have a 'fit' and \
                a 'predict' method.")

        
               